%
% classdef FFTIFO
%
% * Superclass to build IFO by adding optics and progapators
%
%   1) make a file named myIFO.m and define the myIFO as follows
% classdef IFOname < FFTIFO
%     methods
%         function obj = IFOname( varargin )
%             obj@FFTIFO( varargin{:} );
%         end
%         
%         function defineIFO( obj, varargin ) : mandatory to define system
%             % set parameters
%             eval( UtilTK.setVars( who, true, varargin{:} ) );
%             % add optics
%             % connection optics
%             % setup details of optics, like maps
%         end
%         
%         function setupBaseMode( obj )  : optional to define mode
%             call fillQvals to define modes in IFO
%         end
%     end
% end
%
%   2) make an object and lock or calc
%     ifo = myIFO;
%     ifo.lock;  % cavity lengths are adjusted to make ths stationary fields resonant
%     ifo.calc;  % stationary field is calculated with given cavity length
%
%   3) analye
%     fieldName = 'ITM-HR-o';  % 'Optic name' - 'side' - 'i' for in or 'o' for out
%     [E, xy] = ifo.getField(fieldName);  % E(iy,ix) is the complex amplitude at x=xy(ix), y = xy(iy)
%     plot( x, E(end/2,:) )
%     Cmn = ifo.HGCoef(fieldName, 0:5,0:5);    % HG amplitude for m=0~5, n=0~5
%     P = ifo.power(fieldName);                % power of the field
%     Fmn = ifo.HGFrac(fieldName, 0:5,0:5);    % HG power fraction, = abs(Cmn).^2 / power
%     ifo.gaussFit(fieldName)                  % fit of the data by gaussian form
%
% * derived class should implement the following methods
%
%       1) defineIFO : mandatory 
%           This defines the IFO to be simulated
%           call addMirror and addProp to setup the configuration
%           call setHRfiles to define the HR surface details, setTRfiles the transmission and setARfiles the AR surface
%
%       2) setupBaseMode : optional
%           This defines modes in the entire IFO by specifyinf the q value of a field.
%           obj.cavities.fillQvals( portID, qInv ) will start from the given portID with
%           qInv and propagate mode base to all optics.
%           If the mode of the laser is known, just call
%               obj.cavities.fillQvals( obj.optics.findLaser, 1/q of the input to the front mirror );
%
% * To see inforamtions
%       1) ifo.printAll, which calls
%           ifo.fields.printAll
%           ifo.optics.printAll
%           ifo.props.printAll
%           ifo.cavities.printAll
%
%       2) info,printAll('filename') saves the output in 'filename'
%
% * function groups in FFTIFO
%       * "main functions" called in FFTIFO, mainly for internal use
%       * "access functions" to fetch quantities
%       * "add optics" to construct the system and setup details of optics
%       * "Field quantities" to change them
%       * "internal support functions" to implement locking etc
%
% * Things to know
%        1) In principle, don't access data directly from the class properties, 
%           always use access functions defined in this class, like getField.
%           Some properties may be calculated on demand.
%        2) FFTIFO uses adaptive grid sizes, that means, window sizes on mirrors are all different.
%           So, always get proper coordinate by [E,x] = getField
%

%%
% * Recent history
%
% July 1st, 2015
%  Port specification is always in one name, like ITM-HR-i, not a
%  combination of multiple variables, like 'ITM', 'HR', 'i' as separate variables
%
% April 17th, 2015
%  new : setInputType( obj, varargin ), which can be used to change the input field mode
%
% November 4th, 2014
%   updateNfft( obj, varargin ) is chanhged to accept varargin, and Nfft can be used to force Nfft value
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects, 
%   instead of handle of multiple manager handles
%
% October 23, 2014
%   DebugTK is added as obj.debug, and passed to cavities.siblings, which are passed to GuessObj
%
% October 15, 2014
%   roundtripLoss without arguments returns all round trip loss
%
% October 13, 2014
%   use addMirrorGeneric, instead of addMirror2x2 and addMirror4x4
%
% September 19, 2014
%   gaussFit can accpet xList and Eamp, in addition to the field ID
%
% August 29, 2014
%   function [Fld, xl, yl] = getField( fldID, Wout, Lprop )
%   getField has a new argument, Lprop.
%
% August 12, 2014
%   trans and refl add new fields, 'outField-trans' and 'outField-refl'
%   This was done sometime ago. Some analysis codes may need fix for this...
%
% August 5th, 2014
%   auto curvature correction option added to addMirror, removeCurvature.
%   after correction, 1/RoC -> 1/RoC + optDCurv
%
% July 31, 2014
%   initial guess field is default mode field
%
% July 29, 2014
%   lock and calc accepts third argument, which sets the convergence limit of FP arms
%
% July 23, 2014
%   roundtripLoss( fieldID ) did not pickup correct guessField, fixed
%
% July 10, 2014
%   HGFrac and LGFrac added, which are abs(XXCoef).^2/power
%   setRF( RF ) : unit of RF is Hz, instead of MHz
%
% April 3, 2014
%   Initial public release
%

% October 3, 2013
%   modifed as the base class
%
% September 11, 2013
%   initial implementation by Hiro Yamamoto  hiro@ligo.caltech.edu
%

classdef FFTIFO < handle
    
    properties
        ffts
        twids      % full matrix to calculate fields in IFO, E = M E + src
        optics
        props
        fields
        cavities
        debug
    end
    
    % implementation of methods
    methods
        % constructors
        function obj = FFTIFO( varargin )
            % initialize variables
            obj.ffts = FFTTK;
            obj.twids = twiddle( obj );
            obj.fields = fieldManager( obj );
            obj.optics = opticManager( obj );
            obj.props = propagaterManager( obj );
            obj.cavities = cavityManager( obj );
            obj.debug = DebugTK( obj );
            
            % setup IFO coniguration
            obj.defineIFO( varargin{:} );
            
            % find trees and cavities
            obj.analyzeCavity;
            
            % setup base mode. All ports and fields have q value assigned
            % default : use the FP arm, if exists, to define the entire mode
            obj.setupBaseMode( varargin{:} );
            
            % after modes are defined, setup physical dimension
            obj.updateNfft( varargin{:} )
            
            % now all fields have base mode parameters
            obj.setupMisc;
            
            % inject input field
            obj.startFFT;
            
            % calculate powers using scalar field analytic formula
            obj.twiddler;
            
            % calculate something
            obj.calcIt( varargin{:} );
        end
        
        %% main functions called in FFTIFO
        
        function analyzeCavity( obj, root )
            % if root is not defined, use the laser source as the root
            if ~exist('root', 'var')
                root = obj.optics.findLaser;
            end
            obj.cavities.analyzeCavity( root );
        end
        
        % use the FP arm, if exists, to define the entire mode
        function setupBaseMode( obj, varargin )
            % find mode based using closed loop cavity, HR cavities first
            [qInv, root] = obj.cavities.findModeBaseByClosedCavs;
            
            if root == 0
                % there is no closed loop
                disp('mode is not defined, and cannot proceed');
                exit;
            else
                fprintf(1, 'Cavity mode defined using portID %d with q = %g + i * %g\n', root, real(1/qInv), imag(1/qInv) );
                obj.cavities.fillQvals( root, qInv );
            end
        end
        
        function updateNfft( obj, varargin )
        % change Nfft based on optics and field parameters
            Nfft = obj.optics.bookKeeping;
            eval( UtilTK.setVars( {'Nfft'}, true, varargin{:} ) );
            Nlarge = 1024;
            if Nfft > Nlarge
                fprintf(1,'Nfft (%d) is larger than %d, and the simulation may be slow\n', ...
                    Nfft, Nlarge );
                if Nfft > Nlarge * 10
                    reply = input('Do you want to proceed? Y/N [N]:', 's');
                    if isempty(reply) || strcmpi( 'N', reply )
                        error('Nfft too large and quit');
                    end
                end
            end
            obj.ffts.setNfft( Nfft );
        end

        function setupMisc( obj )
            obj.props.setupMisc;
            obj.optics.setupMisc;
            obj.cavities.setupMisc;
        end
        
        function startFFT( obj )
        % function startFFT
        % find the laser source, and set the input field

            % find laser portID and laser opticID
            [portID, laserID] = obj.optics.findLaser;
            
            % fill the inpit field using the default mode
            obj.fields.setDefField( portID, obj.optics.optList(laserID).power );
            
            % let the field touch with input optics
            obj.inject2IFO;
        end
        
        % find alalytic solution
        function twiddler( obj, srcID )
            if ~exist( 'srcID', 'var' )
                srcID = obj.optics.findLaser;
            end
            obj.twids.buildMat;
            obj.twids.calcFields( srcID );
        end
        
        % do whatever you like, to be overriddedn
        function calcIt( obj, varargin )
            for it = 1 : length( obj.cavities.tree )
                obj.cavities.propFields( obj.cavities.tree{it} );
            end
        end
        
        %% access functions
        
        function [Fld, xl, yl] = getField( obj, fldID, Wout, Lprop )
            % [Fld, x, y] = getField( obj, fldID, Wout, Lprop )
            % return the 2D field map (Fld) and coordinates (x, y)
            % of the field fldID.
            % Fld(jy,ix) is the complex amplitude at x(ix), y(jy) in units
            % of sqrt(W)/m.
            % If Wout is specified, the field is calculated in (W x W) space
            % If Lprop is specified, the field is calculated Lprop from the HR side of the source.
            %   If Wout = 0, Wfft is set to be optimal based on the beam size
            %   If Wout = -1, Wfft is set to be the same as the sink.
            % If Lprop = 0, the field on the sink is calculated, i.e., no special operation taken.
            
            % convert name to ID
            fldID = obj.fields.findNamedField( fldID );
            
            if ~exist( 'Wout', 'var' )
                % No Wout meaans no Lprop ...
                % if Wout is not specified, just convert the internal form to standard
                Fld = obj.fields.conv2std( fldID );
                xl = obj.fields.getXlist(fldID);
                yl = xl;
            else
                % Lprop = 0 means do nothing for propgatation distance
                if ~exist( 'Lprop', 'var' )
                    Lprop = 0;
                end
                [Fld, xl, yl] = obj.fields.getField( fldID, Wout, Lprop );
            end
        end
        
        function power = power( obj, fldIDin )
            if ~exist( 'fldIDin', 'var' )
                fldIDin = 1:length(obj.fields.fieldList);
            end
            % convert name to ID
            if ischar( fldIDin )
                fldID = obj.fields.findNamedField( fldIDin );
                power = obj.power( fldID );
            else
                power = zeros(1,length(fldIDin));
                for id = 1 : length(fldIDin)
                    fldID = obj.fields.findNamedField( fldIDin(id) );
                    power(id) = obj.fields.fieldList(fldID).calcPower;
                end
            end
        end
        
        function power = powerInR( obj, fldID, radius )
            fldID = obj.fields.findNamedField( fldID );
            if ~exist('radius', 'var')
                power = obj.power( fdID );
            else
                power = obj.fields.fieldList(fldID).calcSumIn( radius );
            end
        end

        function [info, cwR] = gaussFit( obj, fldIDin, Eamp )
            % [info, cwR] = gaussFit( fldIDin/xList, E )
            % if there is one argument, it is assumed to be a list of field IDs or a field name
            % if there are two arcuments, first argument is the xList and the second argument is a complex field
            %
            % fit the field by gaussian(in x plane) * gaissian(in y plane)
            % info is for text display, so ifp.gaussFit('ITMX-HR-o') will display the fit
            % exp( -(x-x0)^2*(1/wx^2) + i/(2*Rx) )
            %   cwR(1, 2) for x and y
            %   cwR(:).c0 = center of gauss field
            %   cwR(:).w  = the beam width
            %   cwR(:).invRk = k / RoC
            
            if exist('Eamp', 'var')
                % exmplicit complex amplitude given
                % fldIDin is actually an array of coordinates for the field
                [cwR, info] = modeFitData( fldIDin, fldIDin, Eamp );
            else
                if ~exist('fldIDin', 'var')
                    fldIDin = 1:length(obj.fields.fieldList);
                end
                % if field is specied by name, convert it to ID number
                fldIDin = obj.fields.findNamedField( fldIDin );

                info = 'Gauss fit of fields\n';
                for id = 1 : length(fldIDin)
                    [infoOut, cwROut] = obj.fields.fieldList( fldIDin(id) ).gaussFit;
                    if length(fldIDin) == 1
                        info = infoOut;
                        cwR = cwROut;
                    else
                        info = sprintf( '%s\nField %2d) %s', info, fldIDin(id), infoOut );
                        cwR{id} = cwROut;
                    end
                end
            end
        end
        
        function Cmn = HGCoef( obj, fldID, m, n )
            % Cmn = HGCoef( obj, fldID, m, n )
            % calculare the coefficients of HG(m,n) terms

            % convert name to ID
            fldID = obj.fields.findNamedField( fldID );
            Cmn = obj.fields.fieldList(fldID).HGCoef( m, n );
        end
        
        function Pmn = HGFrac( obj, fldID, m, n )
            % Pmn = HGFrac( obj, fldID, m, n )
            % Power fraction of HG(m,n) terms

            fldID = obj.fields.findNamedField( fldID );
            Pmn = abs(obj.fields.fieldList(fldID).HGCoef( m, n )).^2 / obj.power(fldID);
        end

        function Cpm = LGCoef( obj, fldID, p, m )
            % Cpm = LGCoef( fldID, p, m )
            % calculare the coefficients of LG(p,m) terms
            
            % convert name to ID
            fldID = obj.fields.findNamedField( fldID );
            Cpm = obj.fields.fieldList(fldID).LGCoef( p, m );
        end
        
        function Ppm = LGFrac( obj, fldID, p, m )
            % Cpm = LGFrac( fldID, p, m )
            % calculare the fraction of LG power
            
            fldID = obj.fields.findNamedField( fldID );
            Ppm = abs(obj.fields.fieldList(fldID).LGCoef( p, m )).^2 / obj.power(fldID);
        end

        function [E, x] = Non00Amp( obj, fldID, useFit )
        % returns the minor components, i.e., field - gaussFit or field - LGpm/HGmn (type specified at addLaser)
            fldID = obj.fields.findNamedField( fldID );
            [E, x] = obj.getField( fldID );
            if ~exist( 'useFit', 'var' )
                useFit = 1;
            end
            if useFit
                [~, ~, Efit] = modeFitData( x, x, E );
            else
                % amplitude of the field type specified at addLaser, i.e., the main mode
                Efit = obj.fields.fieldAtFID(fldID) * sqrt(obj.power(fldID));
            end
            E = E - Efit;
        end
        
        function loss = roundtripLoss( obj, fldID )
            % if nothing is specified, return all losses
            if nargin == 1
                loss = zeros( 1, length( obj.cavities.guessList )-1 );
                for ig = 1 : length( obj.cavities.guessList )-1
                    loss(ig) = obj.cavities.guessList(ig).roundtripLoss;
                end
                return
            end

            % loss = roundtripLoss( fldID )
            % calculate rount trip loss of a cavity which has fieldID as a part of it
            fldID = obj.fields.findNamedField( fldID );
            loss = NaN;
            for ig = 1 : length( obj.cavities.guessList )-1
                if ~isempty( find(obj.cavities.closedCavs{obj.cavities.guessList(ig).id} == fldID, 1) )
                    loss = obj.cavities.guessList(ig).roundtripLoss;
                    return;
                end
            end
        end
        
        function reflFldID = refl( obj, inFldID )
            % function reflFldID = refl( obj, inFldID )
            % create a new field, which is induced by the reflection of the inField
            
            % incoming field
            inFldID = obj.fields.findNamedField( inFldID );
            % find the port where the reflected field is going out
            % this can have a sum of reflection and transmission
            [outFldID, optID] = obj.optics.findReflPort( inFldID );
            % create a new field which contains reflection componenent only
            % new name is old name + '-refl'
            name = strcat( obj.fields.fieldList(outFldID).name, '-refl');
            reflFldID = obj.fields.newField( name, outFldID );
            % calculate the reflection only field and store in reflFldID
            obj.fields.fieldList(reflFldID).EFFT = obj.optics.refl( inFldID );
            if obj.optics.isOnARSidePortID( inFldID )
                r = obj.optics.optList(optID).rAR;
            else
                r = obj.optics.optList(optID).rHR;
            end
            obj.fields.fieldList(reflFldID).Eamp = r * obj.fields.fieldList(inFldID).Eamp;
            % set the flag
            obj.fields.fieldList(reflFldID).status = 2048;
        end
        
        function transFldID = trans( obj, inFldID )
            % function transFldID = trans( obj, inFldID )
            % create a new field, which is induced by the transmission of the inField
            
            % incoming field
            inFldID = obj.fields.findNamedField( inFldID );
            % find the port where the transmitted field is going out
            % this can have a sum of reflection and transmission
            [outFldID, optID] = obj.optics.findTransPort( inFldID );
            % create a new field which contains reflection componenent only
            % new name is old name + '-trans'
            name = strcat( obj.fields.fieldList(outFldID).name, '-trans');
            transFldID = obj.fields.newField( name, outFldID );
            % calculate the reflection only field and store in reflFldID
            obj.fields.fieldList(transFldID).EFFT = obj.optics.trans( inFldID );
            obj.fields.fieldList(transFldID).Eamp = ...
                obj.optics.optList(optID).t * obj.fields.fieldList(inFldID).Eamp;
            % set the flag
            obj.fields.fieldList(transFldID).status = 2048;
        end
        
        function resetStatus( obj, fldID )
            % no more needed now that calRefl and calTrans add new fields, instead of override
            if exist('fldID', 'var')
                % field is specified, so reste only that
                fldID = obj.fields.findNamedField( fldID );
                obj.fields.fieldList(fldID).status = 0;
            else
                % reset status of all external fields
                obj.cavities.resetFieldStatus;
                obj.cavities.updateFieldStatus;
                obj.inject2IFO;
            end
        end
        
        function [L0, delL] = getPropLeng( obj, propID )
            % [L0, delL] = getPropLeng( obj, propID )
            % cavity length
            [L0, delL] = obj.props.getPropLeng( propID );
        end
        
        %% add optics
        
        % add laser
        function id = addLaser( obj, name, varargin )
            % id = addLaser( 'name', varargin ) for LGpm or HGmn injection
            % add laser source of HG(ind1,ind2) or LG(ind1,ind2)
            %
            % variables names   : help
            % ------------------ ----------------------------------------------------------------
            % power             : laser power [1]
            % lamCR             : CR laser wavelength [1.064e-6]
            % fieldMode         : 'HG' or HermiteGauss and 'LG' for Laguerre Gauss ['HG']
            % ind1              : first index for HG(ind1,ind2) and LG(ind1,ind2) [0]
            % ind2              : second index for HG(ind1,ind2) and LG(ind1,ind2) [0]
            % ------------------ ----------------------------------------------------------------            
            id = obj.optics.addLaser( name, varargin{:} );
        end
        
        % add detector
        function id = addDetector( obj, name )
            id = obj.optics.addDetector( name );
        end
        
        function id = addMirrorGeneric( obj, MirrorObjName, name, varargin )
            % function id = addMirrorGeneric( obj, MirrorObjH, name, varargin )
            % add a mirror object of class MirrorObjName
            if exist(MirrorObjName, 'class') == 8
                % minimal check if the MirrorObjName is a valid class name
                id = obj.optics.addMirrorGeneric( MirrorObjName, name, varargin{:} );
            else
                error('''%s'' is not a valid class name\n', MirrorObjName );
            end
        end

        function id = addMirror( obj, name, varargin )
            % id = addMirror( 'name', 'variableName1', variableValue1, 'variableName2', variableValue2, ... )
            %
            % Add a mirror and return the ID of the mirror.
            % First argument must be the name of the mirror.
            % If variable incAngle is specified, 4x4 mirror (two input fields and two out going fields
            % on each side of the mirror) is added. Otherwise, 2x2 mirror (one input and one output on
            % each side) is added.
            % BS, folding mirrors and MC mirrors are 4x4 mirrors even if angle can be approximated to be 0.
            %
            % variables names   : help
            % ------------------ ----------------------------------------------------------------
            % invRoC            : inverse of the HR side radius of curvature. + for concave [0/m]
            % invRoCAR          : inverse of the AR side radius of curvature. + for concave [0/m]
            % Aperture          : physical size of the mirror [0.34m]
            % Thickness         : thickness of the substrate [0.2m]
            % T                 : power transmisstance [0]
            % Lhr               : power loss on the HR side (Reflectio on HR = 1-T-Lhr) [0]
            % Lar               : power loss on the AR side (Reflectio on HR = 1-T-Lar) [0]
            % resolution        : minimum spatial resolution of the mirror
            % optTheta          : Together with optPhi, the mirror tilt is defined.
            % optPhi            :    pitch : phi=0, theta = pitch,  yaw : phi=pi/2, theta = yaw
            % optX0             : mirror is moved horizontally by this amount
            % optY0             : mirror is moved vertically by this amount
            % removeTilt        : remove tilt using gaussian field (for Mirror2x2)
            % removeCurvature   : remove curvature mismatch using gaussian field (for Mirror2x2), like pseudo RH
            %                       1/RoC -> 1/RoC + mirror.optDCurv
            % incAngle          : incident angle of the field (for BS)
            %
            % nRef              : refractive index (def:nRef(SiO2))
            % isSpherical       : true for spherical mirror, false for Parabolic mirror [true]
            % invFocal          : 1/focal length in the mirror. If maps are defined, it overrides.
            % WfftUser          : If > 0, Wfft for this mirror. If < 0, abs(WfftUser)*beam size is Wfft [0]
            % virtualAperture   : extra outer space in which field is to be calculated [0]
            % ------------------ ----------------------------------------------------------------
            
            if ~ischar( name )
                error('First argument must be the name of the mirror');
            end
            
            % check if the argument format is OK
            if ~UtilTK.validArgList( varargin{:} )
                error('argument needs to be a list of ''name'' and value pairs');
            end
            
            % try to add a mirror2x2
%            id = obj.optics.addMirror2x2( name, varargin{:} );
            id = obj.optics.addMirrorGeneric( 'Mirror2x2Obj', name, varargin{:} );
            if id == 0
                % if failed, add a mirror4x4
%                id = obj.optics.addMirror4x4( name, varargin{:} );
            id = obj.optics.addMirrorGeneric( 'Mirror4x4Obj', name, varargin{:} );

                % id is 0 if this failed
            end
            if id == 0
                error('variable names incorrect for a new mirror specification');
            end
        end
        
        function setHRfiles( obj, mirrorID, varargin )
            % set details of HR surface
            %   use setTRfiles to define the transmission map and setARfiles to define the AR side reflection
            %
            %   setHRfiles( 'mirrorName', 'dataFile', 'formula',
            %               aperture of power fitting (0 for not specifying)
            %               RoC of the power fitting (set to 0 for no fitting),
            %               orientation of the optic (0:no orientation, 1:90 degree, etc),
            %               'variableName1 in formula', initial value of variable1, 'vN2', v2, ... );
            %   example formula : map + tilt*x will add tilt to the data file surface
            %                     map + (x.^2+y.^2)/2*invRoC will add surface curvature change
            %   dataFile can be empty and all can be defined by formula, like loading COMSOL data
            %
            %   use updateMirMaps( 'mirrorName', 'varName1', value1, ... ) to change the formula later
            %
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).HRfiles.addMap( varargin{:} );
        end
        function setTRfiles( obj, mirrorID, varargin )
            % set details of transmission maps
            % see help for setHRfiles
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).TRfiles.addMap( varargin{:} );
        end
        function setARfiles( obj, mirrorID, varargin )
            % set details of AR side reflection maps
            % see help for setHRfiles
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).ARfiles.addMap( varargin{:} );
        end
        
        function setRPYfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).RPYfiles.addMap( varargin{:} );
        end
        function setRSXfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).RSXfiles.addMap( varargin{:} );
        end
        function setTPXfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).TPXfiles.addMap( varargin{:} );
        end
        function setTSYfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).TSYfiles.addMap( varargin{:} );
        end
        
        function updateMirMaps( obj, mirrorID, varargin )
            % updateMirMaps( obj, mirrorID, varargin )
            %
            % change the variables used in the formulas speficying HRfiles, TRfiles and ARfiles
            % E.g., for a formula, map + (x.^ + y.^2)/(2*RoC) for 'ITM',
            %       ifo.updateMirMaps( 'ITM', 'RoC', 100e3 )
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).updateMirMaps( varargin{:} );
        end
        
        function setModMap( obj, mirrorID, modMapH, modMapVar )
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).modMap = modMapH;
            if exist('modMapVar', 'var' )
                obj.optics.optList(mirrorID).modMapVar = modMapVar;
            end
        end
        
        % add propagator
        function id = addOneProp( obj, varargin )
            % id = addOneProp( obj, varargin )
            %   Same syntax as addProp, but adds one way propagator
            id = obj.props.addOneProp( varargin{:} );
        end
        
        function [idf, idb] = addProp( obj, varargin )
            % [idf, idb] = addProp( 'name', source, sink, macroLength, microLength )
            %   Add two proagators, from source to sink and sink to source
            %   macroLength is rounded to the multiple of wave length
            %   source and sink specification can be
            %       one string specifying optic and port, like 'ITM-HR'
            %   example : addProp( 'FPprop', 'ITM-HR', 'ETM-HR', 3994.5);
            [idf, idb] = obj.props.addProp( varargin{:} );
        end
        
        %% Field quantities
        function setRF( obj, RF )
            % setRF( RF )
            % set RF sideband frequency given in Hz
            obj.fields.kSB = 2*pi*RF/ConstantTK.c;
        end
        
        function setInputType( obj, varargin )
            % setInputType( obj, varargin )
            % set the input laser mode
            [fieldMode, ind1, ind2] = obj.fields.getInputType;
            eval( UtilTK.setVars( who, false, varargin{:} ) );

            obj.fields.setInputType( fieldMode, ind1, ind2 );
            obj.cavities.clearAllTemp;
            obj.startFFT;
        end
        
        %% internal support functions

        % arguments for calc and lock
        %
        % varargin{1} : errorLimit [1e-5]
        % varargin{2} : Nprint [1]
        % varargin{3} : FPErrorLim
        % varargin{4} : reset

        function evol = calc( obj, varargin )
            % evol = calc( errorLimit, Nprint, FPErrorLim )
            if nargout > 0
                evol = obj.fieldAction( false, varargin{:} );
            else
                obj.fieldAction( false, varargin{:} );
            end
        end
        
        function evol = lock( obj, varargin )
            % evol = lock( errorLimit, Nprint, FPErrorLim )
            if nargout > 0
                evol = obj.fieldAction( true, varargin{:} );
            else
                obj.fieldAction( true, varargin{:} );
            end
        end
        
        function evol = fieldAction( obj, doLock, varargin )
            evol = zeros(2,0);
            obj.props.propCounter = 0;
            
            Nprint = 1;
            errorLimit = 1e-5;
            if ~isempty( varargin )
                errorLimit = varargin{1};
                if errorLimit <= 0
                    errorLimit = 1e-5;
                end
            end
            if length( varargin ) > 1
                Nprint = varargin{2};
            end
            
            FPErrorLim = 0;
            if length( varargin ) > 2 && varargin{3} > 0
                FPErrorLim = varargin{3};
            end

            if length( varargin ) > 3
                reset = varargin{4};
            else
                reset = 1;
            end
            
            % mark that none has been calculated
            obj.cavities.resetFieldStatus;
            
            % this updates field status of laser
            obj.inject2IFO;
            
            if doLock
                obj.cavities.freezeLength( false );
                % lock when needed
                for ig = 1 : length( obj.cavities.guessList ) - 1
                    obj.cavities.guessList(ig).status = -1;
                end
            else
                obj.cavities.freezeLength( true );
                % don't change
                for ig = 1 : length( obj.cavities.guessList ) - 1
                    obj.cavities.guessList(ig).status = 0;
                end
            end
            
            % if reset, reset all guessFields and pump fields
            if reset
                for ig = 1 : length( obj.cavities.guessList ) - 1
                    guessField = obj.cavities.guessList(ig).guessID;
                    obj.fields.setDefField( guessField, 1 );
                    obj.fields.fieldList(guessField).EFFT = ...
                        obj.fields.fieldList(guessField).EFFT * obj.fields.fieldList(guessField).Eamp;
                    
                    % this should be redundant
                    obj.cavities.guessList(ig).initGuessField( obj.fields.fieldList(guessField).EFFT );
                    
                    obj.cavities.guessList(ig).fillPumpOuts;
                    
                    % a loop which has only one input, which is the end arm
                    if FPErrorLim > 0
                        if size(obj.cavities.guessList(ig).pumpIn,2) == 1
                            obj.cavities.guessList(ig).errorLimit = FPErrorLim;
                        end
                    end
                end
            end
            
            errMax = 10;
            TSTART = tic;
            for iloop = 1 : 1000
                err = 0;
                PowerOld = obj.fields.fieldList( obj.cavities.guessList(1).guessID ).calcPower;
                for ig = 1 : length( obj.cavities.guessList )-1
                    
                    if ig == 1 || ~obj.cavities.guessList(ig).onTheFly
                        err = max( err, obj.cavities.guessList(ig).action(errMax, Nprint) );
                    end
                end
                PowerNew = obj.fields.fieldList( obj.cavities.guessList(1).guessID ).calcPower;
                done = obj.lockDone;
                if done
                    done = false;
                    if abs(PowerNew-PowerOld) < 0.001*PowerNew
                        
                        if errMax <= errorLimit
                            % if errMax (error limit for each field finder) is the user setting
                            if err <= errMax
                                % if the actual error estmation is less than requirement, done
                                % otherwise, keep the loop using the same condition
                                done = true;
                            end
                        else
                            if errMax > 1
                                % if errMax (err for each field finder) is large, make it 0.1
                                errMax = 0.1;
                            else
                                % otherwise, make it 1/10
                                errMax = errMax / 10;
                            end
                        end
                        
%                         if err <= errorLimit
%                             % if the actual error estmation is less than requirement, done
%                             % otherwise, keep the loop using the same condition
%                             done = true;
%                         end
                    end
                end
                evol(1:2,size(evol,2)+1) = [ obj.props.propCounter; err ];
                if done || toc(TSTART) >= 10 || mod(iloop,10) == 0
                    % if more than 5 secods has elapsed or 10 loops
                    if doLock
                        doing = 'full lock';
                    else
                        doing = 'stationary';
                    end
                    fprintf(1, 'Main %s loop %d ... : err = %g, FFT count = %g, power changed from %g to %g\n', ...
                        doing, iloop, err, obj.props.propCounter, PowerOld, PowerNew );
                    TSTART = tic;
                end
                
                if done
                    break;
                end
            end
            
            obj.cavities.updateFieldStatus;
            
        end
        
        % judge is any criteria, except for loop phase, are OK or not, e.g., dark port is dark
        function done = lockDone( obj )
            if obj.cavities.freezeLength
                done = true;
            else
                done = obj.darkPort;
            end
        end
        
        function [done, CD, delL] = darkPort( obj, CDlimit, delLimit )
            
            done = true;
            BSID = obj.optics.findNamedOptic( 'BS' );
            if BSID == 0
                % if there is no BS, that's it
                return;
            end
            toBSX = obj.optics.findPortID( BSID, -3 );
            toBSY = obj.optics.findPortID( BSID, -2 );
            if obj.props.getPropFrom( toBSX ) < 0 || obj.props.getPropFrom( toBSY ) < 0
                % if either port is not connected, that's it
                return;
            end
            
            if ~exist( 'CDlimit', 'var' )
                CDlimit = 10e-6;
            end
            if ~exist( 'delLimit', 'var' )
                delLimit = 1e-12;
            end
            
            REX = obj.optics.refl( toBSX );
            TEY = obj.optics.trans( toBSY );
            PREX = sum(sum( REX .* conj(REX) ));
            PTEY = sum(sum( TEY .* conj(TEY) ));
            REXTEY = sum(sum( REX .* conj(TEY) ));
            
            Pee = PREX + PTEY;
            Ree = 2*real(REXTEY);
            Iee = 2*imag(REXTEY);
            
            delL = 0;
            CD = ( Pee + Ree ) / ( Pee - Ree );
            
            if CD > CDlimit
                % dark port power = Pee + Ree cos(2 phi) - Iee sin(2 phi)
                % = Pee + Aee cos( 2 phi + atan(Iee/Ree) )
                % Aee = sqrt( Ree^2 + Iee^2 )
                % 2 phi + atan(Iee/Ree) = pi, phi = pi/2 - atan2(Iee, Ree)/2
                phi = mod( pi/2 - atan2(Iee, Ree)/2, pi );
                if phi > pi/2
                    phi = phi - pi;
                end
                delL = phi / (obj.fields.kCR + obj.fields.kSB) / 2;
                
                if abs(delL) > delLimit
                    Edark = REX*exp(1i*phi) + TEY*exp(-1i*phi);
                    Pdark = sum(sum(Edark .* conj(Edark)));
                    CDnew = Pdark / (2 * Pee );
                    
                    if CDnew < CD
                        CD = CDnew;
                        [~, propX1ID] = obj.props.getPropFrom( toBSX );
                        propX2ID = obj.props.findReverseProp( propX1ID );
                        
                        [~, propY1ID] = obj.props.getPropFrom( toBSY );
                        propY2ID = obj.props.findReverseProp( propY1ID );
                        
                        obj.props.propList(propX1ID).delL = obj.props.propList(propX1ID).delL - delL;
                        obj.props.propList(propX2ID).delL = obj.props.propList(propX2ID).delL - delL;
                        obj.props.propList(propY1ID).delL = obj.props.propList(propY1ID).delL + delL;
                        obj.props.propList(propY2ID).delL = obj.props.propList(propY2ID).delL + delL;
                    else
                        delL = 0;
                    end
                end
            end
            
            done = abs(delL) < delLimit || CD < CDlimit;
        end
        
        % calculate field in a loop defined by pathLoop
        % if lockIt, propagator propID is adjuted to lock the cavity
        % calcuation information printed every printIt trial. 0 is no print
        function [err, numTry] = calcOne( obj, lockIt, printIt )
            if ~exist( 'lockIt', 'var' )
                lockIt = false;
            end
            if ~exist( 'printIt', 'var' )
                printIt = 0;
            end
            obj.inject2IFO;
            % calculate firlds in a cavity pathLoop,
            % optionally locking by adjuting propID propergator
            [err, numTry] = obj.cavities.compList(1).calc( lockIt, printIt );
        end
        
        % just lock
        function [err, numTry] = lockOne( obj )
            [err, numTry] = obj.calcOne( true );
        end
        
        function printAll( obj, fid )
            % function printAll( fid )
            % print information about fields, optics, etc.
            % If fid is a string, a file is created and the output is saved in the file.
            % If fid is not string, it is used as the file id to be used in fprintf
            
            fileOpened = false;
            if exist( 'fid', 'var' )
                % if fid is a text, it is assumed to be a file name to store
                if ischar( fid )
                    fid = fopen(fid, 'w');
                    fileOpened = true;
                end
            else
                % if nothing specified, it is 1 for screen
                fid = 1;
            end
            
            % print all information
            obj.fields.printAll( fid );
            obj.optics.printAll( fid );
            obj.props.printAll( fid );
            obj.cavities.printAll( fid );
            
            % if the file is opened here, close it
            if fileOpened
                fclose(fid);
            end
        end
        
        % from the field srcID, propagate and go through the first optic
        function inject2IFO( obj )
            srcID = obj.optics.findLaser;
            [sinkID, propID] = obj.props.getPropTo( srcID, true );
            if sinkID > 0
                obj.props.propField(propID);
                if obj.optics.findTransPort( sinkID ) > 0
                    obj.optics.trans( sinkID );
                elseif obj.optics.findReflPort( sinkID ) > 0
                    obj.optics.refl( sinkID );
                end
                
                % laser and sink belongs to system wide guessList
                loopID = length( obj.cavities.guessList );
                obj.fields.markFieldStatus( srcID, loopID );
                obj.fields.markFieldStatus( sinkID, loopID );
            end
        end
        
    end
    
    methods ( Abstract )
        defineIFO( obj, varargin )
    end
    
    methods ( Static )
    end
    
end
