%
% FieldObj : field object
%
% September 19, 2014
%   gaussFit info are processed in modeFitData
%   calcSumIn uses UtilTK.sumIn, instead of FFTTK
%
% September 16, 2014
%   static member function added to convert from internal to standard
%       Estd = conv2stdSTATIC( Eint, phiCorr, rsq, Wfft )
%
% March 8th, 2014
%   changed the superclass of FieldObj from handle to matlab.mixin.Copyable
%   to allow copying existing fields
%
% March 3rd, 2014 HY
%   Analytic solution, Eamp, is moved from twiddle object to this FieldObj.
%
% September 14, 2013
% 
% ff(1:numFlds) = struct( 'mirID', 0, 'port', 0, 'RoC', 1e18, 'CL', 0, ...
%    'w', 1, 'di', 0, 'Cw', 1, 'fac', 1, 'L', 1,  'CLa', 1, ...
%    'rmapC', 0, 'tmapC', 0, 'transToID', 0, 'temp', 0 );
%
%       mirID : mirror this field is attached to
%       port : port index, >=0 HR, < 0 AR, even in field, odd out field
%       RoC : mirror surface RoC
%       CL = 1 / field curvature, 1/RoC if wave front = mirror curvature
%       w  : beam size on the mirror
%       
%
%
classdef FieldObj < matlab.mixin.Copyable
    
    % Using invQ and PrivateInvQ to implement setInvQ is explained in
    % "Avoiding Property Initialization Order Dependency"
    properties (Dependent)
        % modal model base
        % this is to set invQ dependent variables in invQ set function
        invQ % q = z + i z0; invQ = invRoC - i 2/k 1/w^2
    end
    
    properties(Access=private)
        PrivateInvQ
    end
   
    properties
        % IFO object
        ifo
        
        % name of the object
        name

        % calculated in set.invQ
        % values based on the nominal mode base
        z
        z0
        w
        invRoC

        % phiCorr is generalized invRoC
        % E(standard form) = E(internal form) * exp( -i k/2 r^2 phiCorr ) / Wfft
        phiCorr = 0;
        
        % FFT field in internal form. Use above formula or conv2stdSTATIC to convet to standard format
        % This points to an element of fieldManager.Amp defined in get.EFFT
        EFFT
        
        % optics this field is attached to
        optID
        
        % 'status' is a flag if this has been calculated - may carry some information in the future
        % This is cleared at the entrance of calc and lock calls, and fields in closed loops
        % are filled by non-zeros marking these have valid calculated field. 
        % When getField is called, if this flag is 0, it is traced back to non-zero 
        % fields to calculate.
        status = 0;
        
        % field amplitude calculated by twiddle
        % make this to 1 if twiddle initialization is to be removed.
        % Then, the printInfo will need to be changed.
        Eamp = nan;
    end
   
    methods
        
        function obj = FieldObj( name, ifo )
            obj.ifo = ifo;

            obj.name = name;
            obj.invQ = NaN;
        end

        function xList = getXlist( obj )
            xList = obj.ifo.optics.getXlist( obj.optID );
        end
        
        function Wfft = getWfft( obj )
            Wfft = obj.ifo.optics.getWfft( obj.optID );
        end
        
        function val = get.invQ( obj )
            val = obj.PrivateInvQ;
        end
        
        function set.invQ( obj, newInvQ )
            % set.invQ( newInvQ )
            % sets invQ and calculate various modal model quantities
            obj.PrivateInvQ = newInvQ;
            
            obj.w = MMTK.qinv2W( newInvQ, obj.ifo.fields.kCR );
            obj.invRoC = real( newInvQ );
            obj.z = real( 1/newInvQ );
            obj.z0 = imag( 1/newInvQ );
        end
        
        function updateField( obj, fieldID, props, optics )
            if obj.status > 0
                % this field is ready
                return
            end
            
            [srcID, propID] = props.getPropFrom( fieldID );
            if srcID < 0
                % this is a start of a prop, so find interaction
                tempEFFT = obj.ifo.ffts.zeroMat;
                srcID = optics.findReflPort( fieldID, true );
                if srcID > 0
                    % there is relection
                    % make the input to the relfection input port ready
                    obj.ifo.fields.fieldList(srcID).updateField( srcID, props, optics );
                    % save refletion in tempEFFT
                    tempEFFT = optics.refl( srcID );
                end
                srcID = optics.findTransPort( fieldID, true );
                if srcID > 0
                    % there is transmission
                    % make the input to the tranmission input port ready
                    obj.ifo.fields.fieldList(srcID).updateField( srcID, props, optics );
                    % save the total to tempEFFT
                    tempEFFT = tempEFFT + optics.trans( srcID );
                end
                obj.ifo.fields.fieldList(fieldID).EFFT = tempEFFT;
            else
                % this is at the end of propagator, so let the start ready and fill
                obj.ifo.fields.fieldList(srcID).updateField( srcID, props, optics );
                % let it prop from src to sink
                props.propField( propID );
            end
            % mark it is done
            obj.status = 1024;
        end
        
        % standard format HG function with the mode defined at port fldID
        function [E, xList] = HGField2DatFID( obj, m, n )
            [invRoCQ, wQ] = MMTK.qinv2RinvW( obj.invQ, obj.ifo.fields.kCR );
            xList = obj.getXlist;
            E = HGTK.HGField2DRW( m, n, xList, xList, invRoCQ, wQ, obj.ifo.fields.kCR );
        end

        % standard format HG function with the mode defined at port fldID
        function [E, xList] = LGFieldatFID( obj, p, m )
            [invRoCQ, wQ] = MMTK.qinv2RinvW( obj.invQ, obj.ifo.fields.kCR );
            xList = obj.getXlist;
            E = LGTK.LGFieldRW( p, m, xList, xList, invRoCQ, wQ, obj.ifo.fields.kCR );
        end

        function Cmn = HGCoef( obj, m, n, Fld )
            %  HG coefficient at fldID port, for a standard field Fld
            Cmn = zeros( length(m), length(n) );
            if ~exist('Fld', 'var')
                Fld = obj.conv2std( obj.EFFT );
            end
            for im = 1 : length(m)
                for in = 1 : length(n)
                    Cmn(im,in) = obj.ifo.ffts.calcProdInt( Fld, obj.HGField2DatFID( m(im), n(in) ), obj.getWfft );
                end
            end
        end

        function Cpm = LGCoef( obj, p, m, Fld )
            %  LG coefficient at fldID port, for a standard field Fld
            Cpm = zeros( length(p), length(m) );
            if ~exist('Fld', 'var')
                Fld = obj.conv2std( obj.EFFT );
            end
            for ip = 1 : length(p)
                for im = 1 : length(m)
                    Cpm(ip,im) = obj.ifo.ffts.calcProdInt( Fld, obj.LGFieldatFID( p(ip), m(im) ), obj.getWfft );
                end
            end
        end

        function [info, cwR] = gaussFit( obj, message )
            % [info, cwR] = gaussFit( obj, message )
            % Fit the field by gaussian form
            % exp( -(x-x0)^2*(1/wx^2) + i/(2*Rx) )
            % cwR(1, 2) for x and y
            % cwR(:).c0 = center of gauss field
            % cwR(:).w  = the beam width
            % cwR(:).invRk = k / RoC
            if isempty( obj.EFFT )
                info = 'no field';
            else
                [cwR, info] = modeFitData( obj.getXlist, obj.getXlist, obj.conv2std(obj.EFFT), obj.ifo.fields.kCR );
            end
            if exist( 'message', 'var' )
                info = strcat(message, info);
            end
        end
        
        % function power = calcSumIn( obj, radius )
        %   calcute power in a given radius
        function power = calcSumIn( obj, radius )
%        	power = obj.ifo.ffts.calcSumIn( obj.EFFT, radius/obj.ifo.optics.optList(obj.optID).Wfft );
            % magnitude of the field internal format is standard format * Wfft
            if isempty( obj.EFFT )
                power = 0;
            else
                power = UtilTK.sumIn( obj.getXlist, obj.EFFT / obj.getWfft, radius );
            end
        end
        
        function power = calcPower( obj, stdFld )
            % calculate power of the field
            % |EFFT| = standard amplitude * Wfft
            % integral of standard amp^2 = prodInt( stdamp, stdamp, Wfft )
            % = prodInt( stdamp*Wfft, stdamp*Wfft, 1 ) = prodInt( EFFT, EFFT, 1 )
            if exist('stdFld', 'var')
                power = obj.ifo.ffts.calcProdInt( stdFld, stdFld, obj.getWfft );
            else
                power = obj.ifo.ffts.calcProdInt( obj.EFFT, obj.EFFT, 1 );
            end
        end
        
        function Estd = conv2std( obj, Eint )
        % convert internal format to standard
            if ~exist( 'Eint', 'var' )
                Eint = obj.EFFT;
            end
            if isempty( Eint )
                fprintf(1, 'Field ''%s'' is empty\n', obj.name);
                Estd = [];
            else
                Estd = obj.conv2stdSTATIC( Eint, obj.phiCorr, obj.ifo.optics.getRsq(obj.optID), ...
                    obj.getWfft, obj.ifo.fields.kCR );
            end
        end

        function info = printInfo( obj, all )
            if ~exist( 'all', 'var' )
                all = true;
            end
            if isempty( obj.EFFT )
                if ~all
                    info = '';
                    return;
                end
                str = 'FFT field is empty';
            else
                str = sprintf( 'Power = %g', obj.calcPower );
            end
            if isnan( obj.Eamp )
                analStr = 'no analytic evaluation';
            else
                if obj.Eamp == 0
                    analStr = 'E(analytic) = 0';
                 elseif imag(obj.Eamp) == 0
                 	analStr = sprintf( 'E(analytic) = %g', real(obj.Eamp) );
                else
                    analStr = sprintf( 'E(analytic) = %g + i %g', real(obj.Eamp), imag(obj.Eamp) );
                end
            end
            info = sprintf('name: %s   ====   %s, %s, [status = %d]', obj.name, str, analStr, obj.status );
                
            info = sprintf('%s\nRoC = %g, w = %g, z = %g, z0 = %g, 1/phiCorr = %g', ...
                info, 1/obj.invRoC, obj.w, obj.z, obj.z0, 1/obj.phiCorr );
            info = sprintf('%s\nq = %g + i %g, 1/q = %g - i %g', ...
                info, real(1/obj.invQ), imag(1/obj.invQ), real(obj.invQ), abs(imag(obj.invQ)) );
            
        end

    end
    
    methods (Static)
        function Estd = conv2stdSTATIC( Eint, phiCorr, rsq, Wfft, kCR )
            % conversion from interal to standard format
            Estd = exp( -1i * kCR / 2 * phiCorr * rsq ) .* Eint / Wfft;
        end
    end
    
end