%
% GuessObj : information necessary for guess field calculation
%       This is build after cavity is analyzed and FFT are ready to be used
%       When ths object is created and the path is analyzed,
%       input fields may not be ready to be used
%

%% pathRec = struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 )
%    src field ID
%    action : (1) propagation or    <=
%             (2) reflection or     <=
%             (3) transmision       <=
%             (4) add pump field    <=
%             (5) fill guess field  <=
%             (6) save guess field  <=
%             (7) save pumpOut for others
%     id :
%       prop : propID
%       reflection :
%       transmission :
%     add pump field : id is the GuessObj index from which the field is provided
%           field(srcID) = field(srcID) + dumpObj( id )
%     status :  N : add to exiting field N
%            :  0 : fill anew
%            : -N : store in pump field(N)
%
% October 23, 2014
%   Non static DebugTK is accessed using obj.ifo.debug handle, mostly for debug information printout
%
% July 31, 2014
%   pump field is default mode
%
% November 21th, 2013  Hiro Yamamoto
%

%%
classdef GuessObj < handle
    
    properties
        name = 'unnamed'
        
        id  % index of this guess obj, correspnding to closedCavs(id)
        
        % example : laser (field ID 2) -> PRM HR (field ID 5)
        
        % pumpOut is to be used as sources for others
        % pumpOut{1,:} is the source to produce the pump field, like laser source ( 2 for the example )
        % pumpOut{2,:} is the portID to prepare for others, like field inside of PRM ( 5 for the example )
        % pumpOut{3,:} is the NxN field amp other pump field uses : field of 5 with a given input 2
        % in obj.fetchPumpField( id ), pumpID = find( pumpOut{2,:} == id ); field = pumpOut{3,pumpID}
        pumpOut
        
        % pumpIn is a list of pumpIn sources provided by others
        % pumpIn(1,:) : pump source guess ID (guess ID where this pump field is provided)
        % pumpIn(2,:) : external source ID  ( 2 for the example )
        % pumpIn(3,:) : my port to which this pump field is to be added ( 5 for the example )
        pumpIn
        
        % parameters for the accelaration code
        Nacc = 2;
        Nave = 0;
        Nprint = 5;
        lockPropID = -1;  % first prop connected to the guess field
        
        % handle of parent
        ifo
        
        % list of paths for this loop, part of the tree
        pathList
        
        % reassembled actions by comining all elements in pathList
        % each item in this list is executed one after another for one loop execution
        masterPath
        
        % guessFieldID
        guessID
        % saved guess field
        % at the beginning of the loop, field(guessID) is saved in guessField
        % at the end of the loop, field(guessID) has the new field, guessField has the old field
        guessField
        
        % if this is ON, action for this guess field is called when its inputs are ready
        onTheFly = false;
        
        errorLimit = 1e-6;
        
        % round trip phase is tried to be set to this
        lockPhase = 0;
        
        % status
        % 0 : calc
        % -1 : calc, but for the first calc, lock it. After lock, this becomes 1
        %  1 : lock if not freezed
        %  2 : force lock
        status
    end
    
    methods
        % constructor
        function obj = GuessObj( name, id, ifo )
            obj.id = id;
            obj.ifo = ifo;
            
            % this is a special case, maintained separately, like laser ource input
            if id == 0
                obj.guessID = 0;
                obj.onTheFly = false;
            else
                % first field of the close cavity
                obj.guessID = obj.ifo.cavities.closedCavs{id}(1);
                % if the cavity is a closed cavity all facing HR side, like FP or MC, do it right away
                obj.onTheFly = obj.ifo.cavities.isHRloop( obj.ifo.cavities.closedCavs{id} );
            end
            
            if obj.onTheFly
                obj.Nprint = 20;
            end
            
            if isempty( name )
                if obj.onTheFly
                    numMirrors = length(obj.ifo.cavities.closedCavs{id})-1;
                    switch numMirrors
                        case 2
                            guessName = 'FP';
                        case 3
                            guessName = 'MCTri';
                        otherwise
                            guessName = sprintf('MC%dmir', numMirrors );
                    end
                else
                    guessName = 'guess';
                end
                obj.name = sprintf( '%s%d', guessName, id );
            else
                obj.name = name;
            end
            
            obj.pumpOut = cell(3,0);
            obj.pumpIn = zeros(3,0);
            
            obj.pathList = cell(1,0);
            
            obj.initGuessField;
            
            obj.status = 0;
        end
        
        function info = printInfo( obj )
            info = sprintf('[%2d] ''%s'' : guessID = %3d\nNacc = %2d, Nave = %2d, Nprint = %2d\n', ...
                obj.id, obj.name, obj.guessID, obj.Nacc, obj.Nave, obj.Nprint);
            info = sprintf('%sonTheFlay = %d, errorLimit = %g, lockPhase = %g, status = %d', ...
                info, obj.onTheFly, obj.errorLimit, obj.lockPhase, obj.status);
        end
        
        function clearTempFields( obj )
            for ipo = 1 : size(obj.pumpOut,2)
                obj.pumpOut{3,ipo} = [];
            end
            obj.initGuessField;
        end
        
        % calculate something
        function err = action( obj, errorLimit, Nprint )
            if ~exist( 'errorLimit', 'var' )
                errorLimit = obj.errorLimit;
            end
            if ~exist( 'Nprint', 'var' )
                Nprint = obj.Nprint;
            end
            
            % if freezeLength, then don't lock
            % calc, but lock once
            if obj.status == -1
                lockProp = obj.lockPropID;
                obj.status = 1;  % lock if not frozen
            else
                if obj.ifo.cavities.freezeLength
                    lockProp = 0;
                else
                    lockProp = obj.lockPropID;
                end
            end
            obj.ifo.debug.print(['%s >> guessList( %s ).providePump( %s ) called for %s using src field %s \n' ...
                '    calling %s.AccClosedCavityStationary\n'],  ...
                obj.name, ...
                obj.ifo.cavities.guessList( obj.pumpIn(1,1) ).name , obj.ifo.fields.findName(obj.guessID), ...
                obj.ifo.fields.findName(obj.pumpIn(3,1)), obj.ifo.fields.findName(obj.pumpIn(2,1)), ...
                obj.name );
            err = obj.AccClosedCavityStationary( obj.ifo.cavities.guessList( obj.pumpIn(1,1) ).providePump(obj.guessID) , ...
                errorLimit, lockProp, obj.Nacc, obj.Nave, Nprint );
        end
        
        % new single path is added to this database
        function addPath( obj, path )
            if ~isempty( path )
                obj.pathList{ length( obj.pathList) + 1 } = path;
            end
        end
        
        % set the guess field or empty for now
        % this needs to be called when field calculation starts, at least the first one
        function initGuessField( obj, inField )
            if exist( 'inField', 'var' )
                obj.guessField = inField;
            else
                obj.guessField = obj.ifo.cavities.ifo.ffts.zeroMat;
            end
        end
        
        %% building masterPath
        %
        % after FFT is ready, this is called to convert a sequence of path to
        % a sequence of events -
        % transmission, reflection and propagation
        % pumping into from other loops and
        % pump out preparation for other loops
        %
        function buildMasterPath( obj )
            % analy path is an intermediate data
            % analPath{1,n) is source to analPath(2,n)
            analPath = zeros(3, length( obj.pathList{1} )-1);
            for in = 1 : length( obj.pathList{1} )-1
                analPath(1, in) = obj.pathList{1}(in);
                analPath(2, in) = obj.pathList{1}(in+1);
            end
            for ip = 2 : length( obj.pathList )
                thisPath = obj.pathList{ip};
                thisLeng = length(thisPath);
                for in = 1 : thisLeng-1
                    % find where mismatch occurs
                    analPathLeng = size( analPath, 2 );
                    if (analPath(1, in) ~= thisPath(in)) || ...
                            (analPath(2, in) ~= thisPath(in+1))
                        lengMove = length(thisPath) - in;
                        analPath(:, in+lengMove:analPathLeng+lengMove) = analPath(:, in:analPathLeng);
                        for ik = 0 : lengMove-1
                            analPath(1, in+ik) = thisPath(in+ik);
                            analPath(2, in+ik) = thisPath(in+ik+1);
                        end
                        break;
                    end
                    if analPathLeng ~= size( analPath, 2 )
                        break;
                    end
                end
            end
            
            % sort analPath
            % [-N,M] : external pump source, to be added to M
            % [-N,-M] : discard
            % [N,-M] : pump souce for other guess loop
            % [N,M] : prop
            in = 1;
            while in <= size( analPath, 2)
                keepIt = true;
                if analPath(1, in) < 0 && analPath(2, in) > 0
                    % external pump source, to be added
                    pumpInID = size(obj.pumpIn,2) + 1;
                    obj.pumpIn(2,pumpInID) = -analPath(1, in);
                    ipo = obj.ifo.cavities.findClosedCavID( obj.pumpIn(2,pumpInID) );
                    if ipo > 0
                        obj.pumpIn(1,pumpInID) = ipo;
                    else
                        obj.pumpIn(1,pumpInID) = length( obj.ifo.cavities.guessList );
                        obj.ifo.cavities.guessList(end).addPumpOut( -analPath(1, in), analPath(2, in) );
                    end
                    obj.pumpIn(3,pumpInID) = analPath(2, in);
                    keepIt = false;
                elseif analPath(1, in) < 0 && analPath(2, in) < 0
                    % for now not necessary
                    keepIt = false;
                elseif analPath(1, in) > 0 && analPath(2, in) < 0
                    % pump souce for other guess loop
                    obj.addPumpOut( analPath(1, in), -analPath(2, in) );
                    keepIt = true;
                end
                if keepIt
                    in = in + 1;
                else
                    analPath(:, in:end-1) = analPath(:, in+1:end);
                    analPath(:, end) = [];
                end
            end
            
            for ip = 1 : size( analPath, 2 )
                if analPath(2,ip) == obj.guessID || ~isempty( find( analPath(2,1:ip-1) == analPath(2,ip), 1 ) )
                    % guess field is filled at the beginning, so external source is to be added
                    % if already filled, add it
                    analPath(3,ip) = analPath(2,ip);
                else
                    % this is the first time, so fill it
                    analPath(3,ip) = 0;
                end
            end
            
            % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % masterPath is defined here
            %
            %   save current guess field to obj.guessField
            %   add pumpField for this guess field, if there is any
            %   go through the path
            %   guess field is calculated at the last action
            %   save pumpOut fields for others
            %
            % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            obj.masterPath = repmat( struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 ), 1, 0 );
            % at first, restore guess field
            %            obj.addOneAction( 5 );
            % save the current field at guessID to guessField
            obj.addOneAction( 6 );
            obj.addPumpIn( obj.guessID );
            
            for ip = 1 : size( analPath, 2 )
                obj.addOneAction( analPath( 3, ip ), analPath( 1, ip ), analPath( 2, ip ) );
            end
            % save guess field
            %            obj.addOneAction( 6 );
            % calculate and store pump field for thers
            obj.addOneAction( 7 );
        end
        
        
        % register pump field which are needed for others
        function numPumps = addPumpOut( obj, srcID, portID )
            numPumps = size( obj.pumpOut, 2 );
            for ip = 1 : numPumps
                if obj.pumpOut{1,ip} == srcID && obj.pumpOut{2,ip} == portID
                    return;
                end
            end
            numPumps = numPumps + 1;
            obj.pumpOut{1,numPumps} = srcID;
            obj.pumpOut{2,numPumps} = portID;
            obj.pumpOut{3,numPumps} = [];
        end
        
        
        %% prepare masterPath
        
        % when nargin == 4, i.e., user provides action, srcID and sinkID, call addOnePair
        % id = 5, 6, 7 for preparing action 5,6,7
        function addOneAction( obj, action, id1, id2 )
            if nargin == 4
                obj.addOnePair( action, id1, id2 );
            else
                if action == 5 || action == 6
                    % save and restore guessField
                    ID = length( obj.masterPath ) + 1;
                    obj.masterPath(ID).action = action;
                    obj.masterPath(ID).srcID = obj.guessID;
                    if action == 5
                        obj.addPumpIn( obj.guessID );
                    end
                    
                elseif action == 7
                    % save pumpOut for others
                    for ipo = 1 : size( obj.pumpOut, 2 )
                        destLoopID = obj.ifo.cavities.findClosedCavID( obj.pumpOut{2,ipo} );
                        if ~obj.ifo.cavities.guessList( destLoopID ).onTheFly
                            % for onTheFly loop, no need to save
                            % because it is saved explicitly before action is called
                            obj.preparePumpOut( ipo );
                        end
                    end
                end
            end
        end
        
        % use Eamp as the pump source amplitude and calculate pump field
        % this is, for example, PRC loop in aLIGO configuration before SRC loop is calculated
        function fillPumpOuts( obj )
            for ipo = 1 : size( obj.pumpOut, 2 )
                src = abs( obj.pumpOut{1,ipo} );  % src field, which propagates to sinkID
                sinkID = obj.ifo.props.getPropTo( src, true );
                % pump field is default mode
                obj.ifo.fields.setDefField( sinkID, 1 );
                obj.ifo.fields.fieldList(sinkID).EFFT = ...
                    obj.ifo.fields.fieldList(sinkID).EFFT * obj.ifo.fields.fieldList(sinkID).Eamp;
                obj.pumpOut{3,ipo} = obj.ifo.optics.trans( sinkID );
            end
        end
        
        % define action to prepare pumpout field
        function preparePumpOut( obj, ipo )
            % src props to sinkID to outID
            src = abs( obj.pumpOut{1,ipo} );
            outID = obj.pumpOut{2,ipo};
            sinkID = obj.ifo.props.getPropTo( src, true );
            % should be trasmitted
            obj.addReflOrTrans( sinkID, outID, -ipo );
        end
        
        function [propTo, propID] = addOneProp( obj, srcID )
            [propTo, propID] = obj.ifo.props.getPropTo( srcID );
            if propID <= 0
                return;
            end
            for ip = 1 : length( obj.masterPath )
                if obj.masterPath(ip).action == 1 && obj.masterPath(ip).id == propID
                    return;
                end
            end
            ID = length( obj.masterPath ) + 1;
            obj.masterPath(ID).srcID = srcID;
            obj.masterPath(ID).action = 1;
            obj.masterPath(ID).id = propID;
        end
        
        function [kind, ID] = addReflOrTrans( obj, srcID, sinkID, statusID )
            kind = 0;
            if sinkID < 0
                kind = -sinkID;  % ????
            else
                if obj.ifo.optics.findReflPort( srcID ) == sinkID
                    kind = 2;
                elseif obj.ifo.optics.findTransPort( srcID ) == sinkID
                    kind = 3;
                end
            end
            if kind == 0
                return;
            end
            ID = length( obj.masterPath ) + 1;
            obj.masterPath(ID).action = kind;
            obj.masterPath(ID).srcID = srcID;
            obj.masterPath(ID).statusID = statusID;
        end
        
        %  N to M : propagate and trans or refl
        % -N to M : case 4 get external pump source
        %  N to -M : call action if onTheFly flag is ON
        function addOnePair( obj, statusID, srcID, sinkID )
            if srcID > 0 && sinkID > 0
                % add prop
                [propTo, propID] = obj.addOneProp( srcID );
                if propID <= 0
                    return;
                end
                
                if sinkID == obj.guessID
                    statusID = 0;
                end
                if obj.addReflOrTrans( propTo, sinkID, statusID ) > 0
                    % if pumpIn is to be added
                    if sinkID ~= obj.guessID
                        obj.addPumpIn( sinkID );
                    end
                end
            elseif srcID < 0 && sinkID > 0
                % external pump field
                % for now, manually add when internal field is calculated
                %{
                ID = length( obj.masterPath ) + 1;
                obj.masterPath(ID).srcID = sinkID;
                obj.masterPath(ID).action = 4;
                obj.masterPath(ID).statusID = statusID;
                obj.masterPath(ID).id = obj.ifo.cavities.findClosedLoop( abs(srcID) );
                if obj.masterPath(ID).id == 0
                    % the last guessList is system wide server
                    obj.masterPath(ID).id = length( obj.ifo.cavities.guessList );
                end
                %}
            elseif srcID > 0 && sinkID < 0
                % go into other guess loop
                [~, propID] = obj.addOneProp( srcID );
                if propID <= 0
                    return;
                end
                iguess = obj.ifo.cavities.findClosedCavID( abs(sinkID) );
                if obj.ifo.cavities.guessList(iguess).onTheFly
                    ipo = find( [obj.pumpOut{2,:}] == abs(sinkID), 1 );
                    obj.preparePumpOut( ipo );
                    ID = length( obj.masterPath ) + 1;
                    obj.masterPath(ID).action = 7;
                    obj.masterPath(ID).srcID = iguess;
                end
            end
        end
        
        % add action to pumpIn
        function addPumpIn( obj, srcID )
            ipi = find( obj.pumpIn(3, :) == srcID, 1 );
            if ~isempty( ipi )
                ID = length( obj.masterPath ) + 1;
                obj.masterPath(ID).srcID = srcID;
                obj.masterPath(ID).action = 4;
                obj.masterPath(ID).statusID = obj.pumpIn(2,ipi); % only non-zero
                obj.masterPath(ID).id = obj.pumpIn(1,ipi);  % pump source guess loop
            end
        end
        
        %% help to execute masterPath
        
        % on request, return pump field for other guess field calculation
        % outID is the field ID to which the pump field is to be added
        function fld = providePump( obj, outID )
            % find the pumpOut index which has the pump field
            ipo = find( [obj.pumpOut{2,:}] == outID );
            obj.ifo.debug.print('==> pumpOut requested to fill field %s, originated by field %s\n', ...
                obj.ifo.fields.findName(outID), obj.ifo.fields.findName(obj.pumpOut{1,ipo}) );
            if isempty( obj.pumpOut{3,ipo} )
                % if the dump field is not ready yet, calculate now
                src = abs( obj.pumpOut{1,ipo} );
                [sinkID, propID] = obj.ifo.props.getPropTo( src, true );
                obj.ifo.props.propField(propID);
                if obj.ifo.optics.findReflPort( sinkID ) == outID
                    obj.pumpOut{3,ipo} = obj.ifo.optics.refl( sinkID );
                    obj.ifo.debug.print('   pumpOut calculated by reflecting field %d\n', sinkID );
                elseif obj.ifo.optics.findTransPort( sinkID ) == outID
                    obj.pumpOut{3,ipo} = obj.ifo.optics.trans( sinkID );
                    obj.ifo.debug.print('   pumpOut calculated by transmissing field %d\n', sinkID );
                else
                    obj.pumpOut{3,ipo} = obj.ifo.cavities.fields.ffts.zeroMat;
                    obj.ifo.debug.print('   pumpOut filled by empty field\n' );
                end
            else
                obj.ifo.debug.print('   pumpOut field already prepared\n' );
            end
            fld = obj.pumpOut{3,ipo};
        end
        
        %% execute the masterPth instruction set
        
        % addGuessPump controls if the pump field for the guess field is to be added or not
        % For calculational convenience, pump field can be explicitly added to the guess field
        % before goMasterPath is called. When added already, make addGuessPump false to
        % avoid dumpling the pumping for the guess field
        function goMasterPath( obj, addGuessPump )
            for istep = 1 : length( obj.masterPath )
                srcID = obj.masterPath(istep).srcID;
                statusID = obj.masterPath(istep).statusID;
                ID = obj.masterPath(istep).id;
                nextAction = obj.masterPath(istep).action;
                switch nextAction
                    case 1   % propagation
                        obj.ifo.props.propField( ID );
                        obj.ifo.debug.print( '%s\n', obj.ifo.props.propInfo(ID) );
                    case 2   % reflection
                        if statusID == 0
                            obj.ifo.optics.refl( srcID );
                            obj.ifo.debug.print( '%s and output is stored\n', obj.ifo.optics.reflInfo( srcID ) );
                        elseif statusID > 0
                            obj.ifo.fields.fieldList( statusID ).EFFT = ...
                                obj.ifo.fields.fieldList( statusID ).EFFT + ...
                                obj.ifo.optics.refl( srcID );
                            obj.ifo.debug.print( '%s\n     and output is added to field %s\n', ...
                                obj.ifo.optics.reflInfo( srcID ), obj.ifo.fields.findName(statusID) );
                        else
                            obj.pumpOut( -statusID ) = ...
                                obj.ifo.optics.refl( srcID );
                            obj.ifo.debug.print( '%s\n     and output is stored in pump field %s\n', ...
                                obj.ifo.optics.reflInfo( srcID ), obj.ifo.fields.findName(obj.pumpOut{2, -statusID}) );
                        end
                    case 3   % transmission
                        if statusID == 0
                            obj.ifo.optics.trans( srcID );
                            obj.ifo.debug.print( '%s and output is stored\n', obj.ifo.optics.transInfo( srcID ) );
                        elseif statusID > 0
                            obj.ifo.fields.fieldList( statusID) = ...
                                obj.ifo.fields.fieldList( statusID ) + ...
                                obj.ifo.optics.trans( srcID );
                            obj.ifo.debug.print( '%s\n     and output is added to field %s\n', ...
                                obj.ifo.optics.transInfo( srcID ), obj.ifo.fields.findName(statusID) );
                        else
                            obj.pumpOut{ 3, -statusID } = ...
                                obj.ifo.optics.trans( srcID );
                            obj.ifo.debug.print( '%s\n     and output is stored in pump field %s\n', ...
                                obj.ifo.optics.transInfo( srcID ), obj.ifo.fields.findName(obj.pumpOut{2, -statusID}) );
                        end
                    case 4  % add pump field
                        if addGuessPump || (srcID ~= obj.guessID)
                            % if dont add pump field for guess field, then dont add pump for guess
                            if statusID == 0
                                obj.ifo.fields.fieldList( srcID ).EFFT = ...
                                    obj.ifo.cavities.guessList(ID).providePump( srcID );
                                obj.ifo.debug.print( 'pump field stored to %s from guess ID %d\n', ...
                                    obj.ifo.fields.findName(srcID), ID );
                            else
                                obj.ifo.fields.fieldList( srcID ).EFFT = ...
                                    obj.ifo.fields.fieldList( srcID ).EFFT + ...
                                    obj.ifo.cavities.guessList(ID).providePump( srcID );
                                obj.ifo.debug.print( 'pump field added to %%s from guess %d\n', ...
                                    obj.ifo.fields.findName(srcID), ID );
                            end
                        end
                    case 5  % restore guessfield
                        obj.ifo.fields.fieldList( obj.guessID ).EFFT = obj.guessField;
                        obj.ifo.debug.print( 'guess field restored to field( %d )\n', obj.guessID );
                        
                    case 6  % save guess field
                        obj.guessField = obj.ifo.fields.fieldList( obj.guessID ).EFFT;
                        obj.ifo.debug.print( 'guess field is saved from field( %s )\n', obj.ifo.fields.findName(obj.guessID) );
                    case 7  % call other guess field
                        obj.ifo.cavities.guessList( srcID ).action;
                        obj.ifo.debug.print( 'action for guess field loop %d called\n' , srcID );
                end
            end
        end
        %% loop calculation
        
        % To use Epump = N for AccClosedCavityStationary and ClosedCavityStationary,
        % guess field needs to be initialized as the pumpIn field
        
        % use accelarated code by Richard
        % AccN : how many iterations for accelaration calculation
        % AveN : how many samples used for one iretation
        % if lockPropID > 0, then this prop is changed to move the cavity to locked point
        function [err, numTry] = ...
                AccClosedCavityStationary( obj, Epump, errorLimit, lockPropID, AccN, AveN, Nprint )
            if ~exist( 'lockPropID', 'var' )
                lockPropID = 0;
            end
            if isa( lockPropID, 'char' )
                nameStr = lockPropID;
                lockPropID = obj.ifo.props.findNamedProp( lockPropID );
                if isempty(lockPropID)
                    error('This propagator does not exist : ''%s''', nameStr);
                end
            end
            if lockPropID == -1
                lockPropID = obj.ifo.props.getPropID( obj.guessID );
            end
            
            if ~exist( 'Nprint', 'var' )
                Nprint = 0;
            end
            if ~exist( 'AccN', 'var' )
                AccN = 5;
            end
            if ~exist( 'AveN', 'var' )
                AveN = 2;
            end
            if AccN <= 1
                % no acceleration, just simple method
                [err, numTry] = obj.ClosedCavityStationary( Epump, errorLimit, lockPropID, Nprint );
                return;
            end
            
            if length( Epump ) == 1
                % is Epump is given by a number,
                % pump field is the value stored in the field(guessID) and
                % the true initial guess field is Empump * the pump field
                if Epump > 0
                    factor = Epump;
                else
                    factor = 1;
                end
                Epump = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                obj.ifo.fields.fieldList(obj.guessID).EFFT = ...
                    factor * obj.ifo.fields.fieldList(obj.guessID).EFFT;
            end
            
            NC = cell(AccN,1);
            NC(:) = {zeros(size(Epump))};
            EN = cell(AccN+1,1);
            EN(:) = {zeros(size(Epump))};
            
            if AveN==0
                AveN=ones(1,AccN);
            else
                AveN=AveN*ones(1,AccN);
                AveN(2:2:end)=AveN(2:2:end)+1;
            end
            
            numTry = 0;
            for iloop=1:1000
                if lockPropID > 0
                    obj.easyLock( lockPropID );
                end
                
                Elast = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                
                EN(:) = {zeros(size(Epump))}; NC(:) = {zeros(size(Epump))};
                for m=1:AccN
                    for l=1:AveN(m)
                        numTry = numTry + 1;
                        eGuess = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                        EN{m} = EN{m} + eGuess;
                        obj.goMasterPath( false );
                        NC{m} = NC{m} + (eGuess - obj.ifo.fields.fieldList(obj.guessID).EFFT);
                        obj.ifo.fields.fieldList(obj.guessID).EFFT = ...
                            obj.ifo.fields.fieldList(obj.guessID).EFFT + Epump;
                        obj.ifo.debug.print('%s >> Epump added to %s\n', obj.name, obj.ifo.fields.findName(obj.guessID));
                        obj.ifo.debug.print('%s >> end of %d th averaging out of %d\n', obj.name, l, AveN(m));
                    end
                    EN{m}=EN{m}/AveN(m); NC{m}=NC{m}/AveN(m);
                    obj.ifo.debug.print('%s >> end of %d th acceleration of out %d\n', obj.name, m, AccN);
                end
                
                Et{1} = Epump;
                EG = R_getguessFin( NC',EN',Et);
                obj.ifo.fields.fieldList(obj.guessID).EFFT = EG{1};
                
                err = UtilTK.evalError( Elast, obj.ifo.fields.fieldList(obj.guessID).EFFT );
                
                 if err < errorLimit
                     obj.goMasterPath( false );  % test convergence
                     obj.ifo.fields.fieldList(obj.guessID).EFFT = EG{1};
                 end
                
                if Nprint > 0 && mod(iloop,Nprint)==0
                    fprintf(1, '    %s >> Acc Stationary finding : %d tries  power = %g, error = %g, FTT count = %g\n', ...
                        obj.name, numTry, obj.ifo.fields.fieldList(obj.guessID).calcPower, err, obj.ifo.props.propCounter  );
                end
                
                if err < errorLimit
                    break;
                end
            end
        end
        
        % stationary field in a closed cavity defined by cavityPath, with input field Epump, is calculated
        % this is no acceleration
        % When Epump is a positive number, Epump is the initial guess field value and the guess field is set to
        % be Epump x guess field. For a FP, set the initial value of guess field to be the pump in field and
        % set the Epump to be finess, and the convergence could be accelerated.
        %
        % if Nprint > 0, convergence information is printed every Nprint
        % if lockPropID > 0, then this prop is changed to move the cavity to locked point
        %
        % This is a copy of ClosedCavityStationary from cavityManager, customized for GuessObj.
        % The one in cavityManager will be removed in the future.
        %
        function [err, numTry] = ClosedCavityStationary( obj, Epump, errorLimit, lockPropID, Nprint )
            if ~exist( 'lockPropID', 'var' )
                % no locking if lockPropID is not given
                lockPropID = 0;
            end
            if isa( lockPropID, 'char' )
                % if lockPropID is given by name, change it to ID
                nameStr = lockPropID;
                lockPropID = obj.ifo.props.findNamedProp( lockPropID );
                if isempty(lockPropID)
                    error('This propagator does not exist : ''%s''', nameStr);
                end
            end
            if lockPropID == -1
                lockPropID = obj.ifo.props.getPropID( obj.guessID );
            end
            if ~exist( 'Nprint', 'var' )
                % default is no printing
                Nprint = 0;
            end
            
            if length( Epump ) == 1
                % is Epump is given by a number,
                % pump field is the value stored in the field(guessID) and
                % the true initial guess field is Empump * the pump field
                if Epump > 0
                    factor = Epump;
                else
                    factor = 1;
                end
                Epump = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                obj.ifo.fields.fieldList(obj.guessID).EFFT = ...
                    factor * obj.ifo.fields.fieldList(obj.guessID).EFFT;
            end
            
            for numTry = 1:2000
                if (lockPropID > 0) && (mod(numTry, 10) == 1)
                    obj.easyLock( lockPropID );
                end
                
                obj.goMasterPath(false);
                obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.ifo.fields.fieldList(obj.guessID).EFFT + Epump;
                err = UtilTK.evalError( obj.guessField, obj.ifo.fields.fieldList(obj.guessID).EFFT );
                 if err < errorLimit
                     obj.goMasterPath( false );  % test convergence
                     obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.ifo.fields.fieldList(obj.guessID).EFFT + Epump;
                 end

                if (Nprint > 0) && (mod(numTry, Nprint) == 0)
                    fprintf(1, '    %s >> Stationary finding : %d  power = %g, error = %g, FFT count = %g\n', ...
                        obj.name, numTry, obj.ifo.fields.fieldList(obj.guessID).calcPower, err, obj.ifo.props.propCounter );
                end
                if err < errorLimit
                    break;
                end
            end
        end
        
        % change the microscopic path length of the propergator propID to make the round trip phase to null
        function delL = easyLock( obj, propID )
            delL = ( obj.calcLoopPhase - obj.lockPhase ) / (obj.ifo.fields.kCR + obj.ifo.fields.kSB);
            backID = obj.ifo.props.findReverseProp( propID );
            if backID > 0
                % this length change is applied for two paths, go and back, so make it half
                delL = delL / 2;
                obj.ifo.props.propList(backID).delL = obj.ifo.props.propList(backID).delL + delL;
            end
            obj.ifo.props.propList(propID).delL = obj.ifo.props.propList(propID).delL + delL;
        end
        
        % find round trip phase
        function loopPhase = calcLoopPhase( obj )
            current = obj.ifo.cavities.freezeLength( true );
            
            % when the masterPath starts, gield(guessID) is stored in obj.guessField
            obj.goMasterPath( false );
            loopPhase = angle( sum(sum( obj.ifo.fields.fieldList(obj.guessID).EFFT .*conj(obj.guessField) )) );
            % restore initial state
            obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.guessField;
            
            obj.ifo.cavities.freezeLength( current );
        end
        
        function loss = roundtripLoss( obj )
            % function loss = roundtripripLoss( obj )
            % find round trip loss
            
            current = obj.ifo.cavities.freezeLength( true );
            
            Pold = obj.ifo.fields.fieldList( obj.guessID ).calcPower;
            % when the masterPath starts, gield(guessID) is stored in obj.guessField
            obj.goMasterPath( false );
            
            Pnew = obj.ifo.fields.fieldList( obj.guessID ).calcPower;
            % at each bounce, reflectance is multiplied
            RProd = 1;
            for im = 1:length( obj.masterPath )
                % for each reflection, multiply reflectance on HR
                if obj.masterPath( im ).action == 2
                    [~, optID] = obj.ifo.optics.isHRside( obj.masterPath( im ).srcID );
                    %{
                    if isHR
                        rProd = rProd *  obj.ifo.optics.optList(optID).rHR;
                    else
                        rProd = rProd *  obj.ifo.optics.optList(optID).rAR;
                    end
                    %}
                    RProd = RProd * obj.ifo.optics.optList(optID).Rhr0;
                end
            end
            loss = 1 - Pnew/RProd/Pold;
            
            % restore initial state
            obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.guessField;
            
            obj.ifo.cavities.freezeLength( current );
        end
        
    end
end
