% Hermite Gauss functions
%
% Septemer 2nd, 2015
%   explicit argument of kCR
%
% October 14, 2013  Hiro Yamamoto  hiro@ligo.caltech.edu
%   first release
%

classdef HGTK < handle
    
    methods (Static)
  
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % 2D Hermite Gaussian with invQ as input
        % HGField2D( m, n, x, y, invQ ) or
        % HGField2DRW( m, n, x, y, invRoC, w )
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
        function E = HGField2D( m, n, x, y, invQ, k )
          	[invRoC, w] = MMTK.qinv2RinvW( invQ, k );
            E = HGTK.HGField2DRW( m, n, x, y, invRoC, w, k );
        end
        
        function E = HGField2DRW( m, n, x, y, invRoC, w, k )
            E = HGTK.HGField1D( n, y, invRoC, w, k ).' * HGTK.HGField1D( m, x, invRoC, w, k );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % hermite gaussian field specified by
        % the inverse of RoC and the beam size
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function E = HGField1D( m, x, invRoC, w, k )
            E = sqrt( sqrt(2)/w ) * HGTK.HG1DN( m, sqrt(2)*x/w ) .* exp(-1i*k*invRoC/2*x.^2 );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % normalized (w=1) 1D Hermite Gaussian
        % int[-inf,inf] HG(m,x) HG(n,x) = 1 for m=n
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function data = HG1DN( m, x )
            if size(m,1) < size(m,2)
                m = m';
            end
            if size(x,1) > size(x,2)
                x = x';
            end
            
            data = ...
                repmat(1 ./ sqrt( sqrt(pi) * pow2(m) .* gamma(m+1) ),1,size(x,2)) ...
                .* HGTK.Hermite( m, x ) ...
                .* repmat(exp( -x.^2/2 ), size(m,1), 1 );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Hermite polynomial
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function data = Hermite( m, x )
            if size(m,1) < size(m,2)
                m = m';
            end
            if size(x,1) > size(x,2)
                x = x';
            end
            xLeng = size(x,2);
            
            data = zeros( length(m), length(x) );
            for im = 1 : length(m)
                switch m(im)
                    case 0,
                        data(im,1:xLeng) = 1;
                    case 1,
                        data(im,:) = 2*x;
                    case 2,
                        data(im,:) = -2 + 4*x.*x;
                    case 3,
                        data(im,:) = ( 8*x.*x - 12 ) .* x;
                    case 4,
                        x2 = x.*x;
                        data(im,:) = (16*x2 - 48).*x2 + 12;
                    case 5,
                        x2 = x.*x;
                        data(im,:) = ((32*x2-160).*x2 + 120).*x;
                    otherwise
                        data(im,1:xLeng) = 1;
                        coeff = 1;
                        ms = 0;
                        s2 = m;
                        x22 = 4*x.*x;
                        for s = 1 : floor(m/2)
                            ms = ms + 1;
                            coeff = -coeff * s2*(s2-1) / ms;
                            s2 = s2 - 2;
                            data(im,:) = data(im,:) .* x22 + coeff;
                        end
                        if mod(m,2) ~= 0
                            data(im,:) = data(im,:) * 2 .* x;
                        end
                end
            end

        end
    end
end
