%
% propagator : connects two ports
%
% September 2nd, 2015
%    ifo is added as a property
%    several functions with optics and fields as input arguments are
%    modified to use obj.ifo.optics and .fields
%
% April 9, 2015
%    the sign of the ratio for the adaption gridding is chosen 
%    so that the extra phase of Amap becomes small
%           Amap = exp( 1i * k/2 * ( (ratio-1)/Lprop - getInvRoC(inPortID) ) * r^2 )
%
% September 18, 2014
%   setupMisc( obj, optics, fields, Lprop, sinkWfft, Nprop )
%   outField = propField( obj, inField, kSB, Nprop )
%   [outField, xl, yl] = propOnDemand( obj, optics, fields, Wsink, Lprop )
%
%   These are updated to support getField( propagation distance and output window size )
%   There is some issue in propOnDemand. Some sink window size does not work well, check out WHYWHYWHY
%
% August 29, 2014
%	function [outField, xl, yl] = propOnDemand( obj, optics, fields, Wout, Lprop )
%   When Lprop > 0, the propagation distance is changed
%   For this case, if Wout == 0, the window size is scaled proportional to Lprop
%
% September 11, 2013  Hiro Yamamoto
%
%
classdef PropObj < handle
    
    properties (SetObservable)
        % macroscopic length, integer multiple of the base wavelength
        L0 = NaN;
        % refractive index
        nRef = 1;
    end
    
    properties
        % microscpic length. Total length = L0 + delL
        delL = 0;
        
        % the extra phase for SB is calculated by exp( i kSB propL0 )
        propL0 = NaN;
        
        % input and output field IDs
        inPortID = NaN;
        outPortID = NaN;
        
        % coordinate scaling
        ratio = 1;  % Wfft(sink) / Wfft(src)  x -1 when the curvature changes sign
        CLa = 0;    % (1 - ratio) / (ratio * L0);
        
        % map to be multipled on field before propagation (A)
        Amap = zeros(0,0);
        
        % FFT Propagation matrix (K)
        Pmap = zeros(0,0);
        
        % name of prop
        name = 'unnamed';
        
        % propagator ABCD
        propABCD
        
        % gouy phase
        gouy00
        
        % lock phase
        lockPhase  % one trip phase for this cavity
        
        % IFO object
        ifo
    end
    
    % function to replace a static variable
    methods
        % constructor
        function obj = PropObj( name, L0, delL, inPortID, outPortID, ifo )
            obj.ifo = ifo;
            
            obj.outPortID = outPortID;
            obj.inPortID = inPortID;
            obj.delL = delL;
            obj.L0 = L0;    % this updates propABCD
            obj.name = name;
            obj.lockPhase = 0;
            
            addlistener(obj, 'nRef', 'PostSet', @obj.updateNdep);
        end
        
        %%	
        %  Propagation using internal format and the conversion from internal to stanfdard format
        %       Eout = ifft( Pmap * fft( Amap * Ein ) ) * exp( -i (k0 * dL + kSB * L0) )
        %       Enominal = Einternal * exp( -i k/2 r^2 phiCorr ) / Wfft(field)
        %
        %       Ein, Eout, Amap : Nfft/2 x Nfft/2 (Nfft is the standard Nfft)
        %       Pmap : Nprop x Nprop (Nprop can be larger than Nfft for larger beam propagation)
        %
        %   setupMisc( obj, optics, fields, Lprop, sinkWfft, Nprop )
        %       calculates Pmap, Amap and phiCorr (stored in field)
        %       sinkWfft = 0  : optimal window size
        %                = -1 : same as the default propagation (good when Lprop ~ L0)
        %
        %   Eout = propField( obj, inField, kSB, Nprop )
        %       Eout = ifft( Pmap * fft( Amap * Ein ) ) * exp( -i (k0 * dL + kSB * L0) )
        %
        %
        %%
        function setupMisc( obj, Lprop, sinkWfft, Nprop )
            % calculate parameters used for field progation
            %
            %   adaptive grid size is used for beam size change, srcWfft and sinkWfft are used for the ratio
            %   If the waist is passed during this propagation, ratio = -ratio, sort of slipping up-down/right-left
            %
            % some parameters for calculating propagation parameter on demand
            %   Lprop - propagation distance, which can be different from the source and sink optic
            %       if not specified, it is calculated from L0, inPort and outPort
            %   sinkWfft - Wfft after the propagation
            %       undefined or < 0 : sinkWfft at the outPort is used
            %   Nprop - Nfft is replaced by this one in propagation to support bigger beam propagation
            %       srcWfft is expanded to srcWfft x Nnew / srcNfft, with outside being zeroed
            %       if not specified, standard Nfft is used
            
            % parameters calculated by this function
            %   ratio
            %   gouy00
            %   CLa = -(1-1/ratio) / Lprop
            %       this will be used to convert internal format Eout to standard format
            %   Amap, Pmap : out = ifft( Pmap * fft( Amap * Ein ) )
            %   If Lprop is not specified, i.e., standard call
            %       phiCorr in field objects at each side of the propagation
            
            onDemandLprop = exist( 'Lprop', 'var' ) && (Lprop > 0);
            onDemandWsink = exist( 'sinkWfft', 'var' ) && (sinkWfft > 0);

            srcOpt = obj.ifo.optics.findOptic( obj.inPortID );
            srcWfft = obj.ifo.optics.optList( srcOpt ).Wfft;
            % if not specified, use the default Nfft
            if ~exist( 'Nprop', 'var' )
                Nprop = obj.ifo.optics.ifo.ffts.Nfft;
            end
            
            % if not specified, use the Wfft at the end of the propagation
            if ~onDemandWsink
                sinkOpt = obj.ifo.optics.findOptic(obj.outPortID);
                sinkWfft = obj.ifo.optics.optList( sinkOpt ).Wfft;
            end
            
            % If Lprop is not specified, propagation distance is between HR surface to HR surface
            obj.propL0 = obj.L0;
            if ~onDemandLprop
                Lprop = obj.L0 + obj.ifo.optics.getDeapth( obj.inPortID ) + obj.ifo.optics.getDeapth( obj.outPortID );
                obj.propL0 = Lprop;                
            end
            
            % the ratio of scaling
            newSrcWfft = Nprop / obj.ifo.ffts.Nfft * srcWfft;
            obj.ratio = sinkWfft / newSrcWfft;
            
%           sign of ratio is arbitrary, but ...
%           out = ifft( Pmap * fft( Amap * Ein ) )
%           Amap = exp( 1i * k/2 * ( (ratio-1)/Lprop - fields.getInvRoC(obj.inPortID) ) * r^2 )
%           so if ratio = 1 + L/R, phase correction is minial              
%           Choose the sign of ratio to be the sign of (1 + Lprop*fields.getInvRoC(obj.inPortID))
            if (1 + Lprop*obj.ifo.fields.getInvRoC(obj.inPortID)) < 0
                obj.ratio = -obj.ratio;
            end
            
            if Lprop == 0
                obj.gouy00 = 0;
                
                % this case should not happen or used
                obj.CLa = NaN;
                
                % when there is no propagation, just fill the RoC
                obj.ifo.fields.fieldList(obj.inPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.inPortID);
                obj.ifo.fields.fieldList(obj.outPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.outPortID);
                
                % these will never be used
                obj.Pmap = 1;
                obj.Amap = 1;
            else
                obj.gouy00 = MMTK.gouyL( 1/obj.ifo.fields.fieldList(obj.inPortID).invQ, Lprop / obj.nRef );
                
                % calculate maps for the FFT propagation
                % outField(x,y) = ifft( Pmap * fft( Amap * inField ) )
                %   inField internal  = Enominal * exp( i k/2 r^2 invRoC ) Wfft(in)
                %   outField internal = Enominal * exp( -i k/2 r^2 CLa ) Wfft(out)
                
                obj.CLa = -(1 - 1/obj.ratio) / Lprop;
                
                % Enominal = Einternal * exp( -i k/2 r^2 phiCorr ) / Wfft(field)
                
                % The FOG defines CL (=invRoc) for the going out field to be opposite to going in field
                % That makes this definition opposite, but this and FOG are consistent
                if (~onDemandLprop) && (~onDemandWsink)
                    obj.ifo.fields.fieldList(obj.inPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.inPortID);
                    obj.ifo.fields.fieldList(obj.outPortID).phiCorr = -obj.CLa;
                end
                
                % Amap is the same size as original, so it can be multiplied with Ein
                % Pmap is Nnew x Nnew, using larger source window, filled with 0 out of the original source window
                % out = ifft( Pmap * fft( Amap * Ein ) )
                obj.Amap = exp( 1i * obj.ifo.fields.kCR/2 * ...
                    ( (obj.ratio-1)/Lprop - obj.ifo.fields.getInvRoC(obj.inPortID) ) * ...
                    obj.ifo.optics.getRsq( srcOpt ) );
                
                Fvec = 1/(2* newSrcWfft)*(-Nprop:2:(Nprop-2));
                [fx,fy] = meshgrid( Fvec, Fvec );
                obj.Pmap = exp( 1i * pi * obj.ifo.fields.lamCR * Lprop / obj.ratio * (fx.^2 + fy.^2) );
                Ncut = obj.FFTFiltC( newSrcWfft, sinkWfft, Lprop, Nprop, obj.ifo.fields.lamCR );
                obj.Pmap(Ncut) = 0;
                obj.Pmap = ifftshift(obj.Pmap);
            end
        end
        
        %%
        function outField = propField( obj, inField, kSB, Nprop )
            %
            % FFT propagation with magnification, was R_diffrC3 in FOG
            %	Eout = ifft( Pmap * fft( Amap * Ein ) ) * exp( -i (k0 * dL + kSB * L0) )
            %
            % inField and outField are in internal format
            %
            %   Nprop x Nprop is used for the FFT propagation in frequemncy space to calculate larger field props
            %   Default is twice of the field size
            %
            
            srcNftt = 2 * length( inField );
            if ~exist( 'Nprop', 'var' )
                Nprop = srcNftt;
            end
            % out = ifft ( Pmap * fft( Amap * inField ) )
            % expand doubled for alias killing zone
            % When Nprop > srcNfft, the input field region is expanded by this much
            % and the internal format ~ Wfft, so inField needs to be multiplied by this much
            myfield = UtilTK.expand2x2( obj.Amap .* inField, Nprop ) * (Nprop / srcNftt);
            Field_FFT = fft2(myfield);
            Field_FFT = obj.Pmap .* Field_FFT;
            outField = ifft2(Field_FFT);
            
            % remove outside zone used through propagation
            outField = UtilTK.useQuarter( outField );
            
            if obj.ratio < 0
                % rot90( field, 2 ) because axis is reversed
                % - because, after the propagation, 1/ratio is multiplied, ratio is signed
                % 1/abs(ratio) = Wfft(in) / Wfft(out), this part is taken care by the normalization
                % but the sign is not. So it is absorbed here.
                outField = -rot90(outField,2);
            end
            
            % add phase change due to sideband frequency and cavity length offset
            % delL * kSB ~ 10^-9 is neglected
            outField = outField * exp( -1i * ( obj.ifo.fields.kCR * obj.delL + kSB * obj.propL0 ) );
        end
        
        %%
        function [outField, xl, yl] = propOnDemand( obj, Wsink, Lprop )
            % let the field propagate by Lprop, with the window size Wout
            % If Lprop = 0, use the default propagation distance
            % If Lprop > 0,
            %   If Wsink < 0, Wfft is set to be the same as the sink.
            %   If Wsink = 0, Wfft is set to be optimal based on the beam size
            % Wsink is visible window size, and to be doubled to suppress aliasing
            
            tempProp = PropObj( 'tempProp', obj.L0, obj.delL, obj.inPortID, obj.outPortID, obj.ifo );

            srcOpt = obj.ifo.optics.findOptic(obj.inPortID);
            srcWfft = obj.ifo.optics.optList( srcOpt ).Wfft;
            sinkOpt = obj.ifo.optics.findOptic(obj.outPortID);
            sinkWfft = obj.ifo.optics.optList( sinkOpt ).Wfft;  % this is full size including alias killer
            
            if ~exist( 'Wsink', 'var' )
                Wsink = 0;  % optimal size based on beam size
            end
            
            if Wsink > 0
                Wsink = Wsink * 2; % make it doubled to be consistent
            elseif Wsink < 0
                % default sinkWfft
                Wsink = sinkWfft;
            end
            
            useOptimalW = Wsink == 0;
            
            if exist( 'Lprop', 'var' ) && (Lprop > 0)
                % finite propagation
                
                % calculate the beam sizes at the src and sink using invQ
                [z,z0] = MMTK.q2zz0( 1/obj.ifo.fields.fieldList(obj.inPortID).invQ );
                wAtSrc = MMTK.qinv2W( 1/( z + 1i*z0 ), obj.ifo.fields.kCR );
                wAtSink = MMTK.qinv2W( 1/( z + Lprop + 1i*z0 ), obj.ifo.fields.kCR );
                
                % the adaptive scaling
                wScale = wAtSink / wAtSrc;
                
                % the optimal output window based on beam size
                Woptimal = obj.ifo.optics.winFac * wAtSink * 2;  % * 2 for alias killer
                
                if useOptimalW
                    % use optimal window size
                    Wsink = Woptimal;
                else
                    % use specified W or optimal W, whichever is larger
                    if Wsink < Woptimal
                        fprintf(1,'Window size (%g) is suggested to be larger than %g\n', Wsink/2, Woptimal/2 );
                    end
                end
                
                % Wsink is the window at the end
                % Wsrc is the window at the src
                % window size at the end and the ratio of the windows are given, so window at the src is
                Wsrc = Wsink / wScale;
                
                % find new Nprop used for the larger beam
                Nprop = obj.ifo.ffts.Nfft;
                Nratio = Wsrc / srcWfft;
                if Nratio <= 1.5  % WHYWHYWHY
                    % if original srcWfft is large enough, just use the original
                    Wsrc = srcWfft;
                else
                    % Wsrc is needs to be larger
                    % calculate new Nfft to make the new visible window size larger than requested
                    % Nprop is smallest of Nfft x 2^something, but larger than Nfft x Nratio
                    Nprop = 2^ceil( log(Nprop * Nratio)/log(2) );
                    % same bin size, so Wsrc/srcWfft = Nprop / Nfft
                    Wsrc = srcWfft * Nprop / obj.ifo.ffts.Nfft;
                end
                % Wsink / Wsrc = wScale <- adaptive scale based on the beam size
                if useOptimalW
                    Wsink = Wsrc * wScale;
                else
                    rrr =  Wsink / (Wsrc * wScale);
                    if rrr > 1.2
                        % WHYWHYWHY
                        fprintf(1, 'The window size requested (%g) is limited to %g\n', Wsink/2, Wsrc * wScale/2 );
                        Wsink = Wsrc * wScale;
                    end
                end
            else
                Lprop = 0;
                if useOptimalW  % make no sense, but in case...
                    Wsink = sinkWfft;
                end
                
                % make the full FFT window to be double of requested
                Nratio = Wsink / sinkWfft;
                Nprop = obj.ifo.ffts.Nfft;
                % calculate new Nfft to make the new visible window size larger than requested
                Nprop = 2^ceil( log(Nprop * Nratio)/log(2) );
                % same bin size, larger Nfft to make the window bigger
                if useOptimalW
                    Wsink = sinkWfft * Nprop / obj.ifo.ffts.Nfft;
                end
            end
            
            tempProp.setupMisc( Lprop, Wsink, Nprop );
            outField = tempProp.propField( obj.ifo.fields.fieldList(obj.inPortID).EFFT, obj.ifo.fields.kSB, Nprop );
            
            % convert internal format to standard format
            % add phase change due to sideband frequency and cavity length offset
            % also, internal format to standard format conversion is done here
            % Nprop/2 data in -Nprop/4 to Nprop/2
            xl = Wsink / Nprop * (-Nprop/4+0.5:Nprop/4-0.5);
            yl = xl;
            [x,y] = meshgrid( xl, yl );
            
            % phiCorr for outField = -CLa = (1 - 1/thisRatio) / Lprop
            outField = FieldObj.conv2stdSTATIC( outField, -tempProp.CLa, x.^2+y.^2, Wsink, obj.ifo.fields.kCR );
        end
        
        %%
        function loss = propLoss( obj )
            % 1 - power in the sink optic / power at the source
            if obj.L0 == 0
                loss = 0;
            else
                optID = obj.ifo.optics.findOptic( obj.outPortID );
                
                obj.ifo.fields.calcIfNecessary( obj.inPortID );
                obj.ifo.fields.calcIfNecessary( obj.outPortID );
                
                Aperture = obj.ifo.optics.optList(optID).Aperture;
                loss = 1 - obj.ifo.fields.fieldList(obj.outPortID).calcSumIn(Aperture/2)  / obj.ifo.fields.fieldList(obj.inPortID).calcPower;
            end
        end
        
        function done = resetDelL( obj )
            % set the delL of the cavity to cancel the gouy phase
            % resonant cond for one prop is -k l n + eta00 = n pi => l = mod(eta00, pi) / (k n)
            % done = true when both ends are connected
            done = obj.inPortID > 0 && obj.outPortID > 0;
            if done
                obj.delL = mod( obj.gouy00 * obj.ifo.fields.getGouyMul, pi) / (obj.ifo.fields.kCR * obj.nRef);
            end
        end
        
        function set.L0( obj, L0 )
            obj.L0 = round( L0 / obj.ifo.fields.lamCR ) * obj.ifo.fields.lamCR;
            obj.updateNdep;
        end
        
        function updateNdep( obj, ~, ~ )
            obj.propABCD = [ 1, obj.L0/obj.nRef; 0, 1 ];
        end
        
        function info = printInfo( obj )
            info = sprintf('name: %s   prop loss = %g\n', obj.name, obj.propLoss );
            info = sprintf('%s[inPort, outPort] = [%d, %d]\n', info, obj.inPortID, obj.outPortID );
            info = sprintf('%sL0 = %g, delL = %g, gouy00 = %g, nRef = %g, ratio = %g, CLa = %g, propL0 = %g\n', ...
                info, obj.L0, obj.delL, obj.gouy00, obj.nRef, obj.ratio, obj.CLa, obj.propL0 );
        end
        
    end
    
    methods (Static)
        % was R_FFTFiltC, copied from FOG
        % cut larger angle scattering to suppress alias
        % while keeping the scattering from a bottom end of mirror to the other mirror
        function Filter = FFTFiltC(wind_length1,wind_length2, L0, Nfft, lambda)
            rhom = Nfft / (2*wind_length1);
            rhoc = wind_length2/(2*L0*lambda);
            Nc = Nfft*rhoc/rhom;
            if(Nc>Nfft)
                Nc=Nfft;
            end
            mask=ones(Nfft,Nfft);
            val = ceil((Nfft-Nc)/2);
            mask(val+1:end-val,val+1:end-val)=0;
            Filter = mask==1;
        end
    end
    
end
