function inArea = SqInCircle( x0, y0, rX, rY )
% function inArea = SqInCircle( x0, y0, rX, rY )
%
% For a square, centered at (x0,y0)
% the area inside of an oval with radius of (rX, rY) centered at (0,0)
% is calculated.
%

if min( size(x0) ) == 1 && min( size(y0) ) == 1
    [x0,y0] = meshgrid( x0, y0 );
end

% if quad symmetry exists, calculate only one quadrant
quadSym = mod(size(x0,2),2) == 0;
if quadSym
    quadSym = isempty( find( x0(1,1:floor(end/2)) ~= -x0(1,end:-1:floor(end/2)+1), 1) ) & ...
        isempty( find( y0(1:floor(end/2),1) ~= -y0(end:-1:floor(end/2)+1, 1), 1 ) );
end

if quadSym
    [x0,y0] = meshgrid( x0(1,1:end/2), y0(1:end/2,1) );
end

dxy = x0(1,2)-x0(1,1);

% use rX and rY as unit
delX = dxy/rX;
delY = dxy/rY;

% this is the area
inArea = ones( size(x0) ) * delX * delY;

% effectively, flip all to positive side
x0 = abs(x0);
y0 = abs(y0);

% bottom-left corner - closest to the origine
xmin = (x0 - dxy / 2) / rX;
ymin = (y0 - dxy / 2) / rY;

% if the nearest point is out of the region, no points are in the region
inArea( xmin.^2 + ymin.^2 > 1 ) = 0;

% top-right corner - furthest from the origine
% by using min( x, rMax ), the edge effect is automatically taken care
% when the area of ([xmin, xmax], [ymin, ymax]) in a unit circle is calculated.
xmax = xmin + delX;
ymax = ymin + delY;

% if the furthest point is in the region, all are in the region
% so calculate only when the rectangle region is partially included by
% a circle with radius of 1 (normalized by rX and rY)

% nearest point is in, furthest point is out
% so partially in the oval region
onBoundary = find( (xmax.^2 + ymax.^2 >= 1) & (xmin.^2 + ymin.^2 <= 1) );
    
for id = 1 : length(onBoundary)
    index = onBoundary(id);
    topLeftIn = xmin(index)^2 + ymax(index)^2 <= 1;
    botRightIn = xmax(index)^2 + ymin(index)^2 <= 1;
    
    if topLeftIn && botRightIn %% area[11]
        v1 = ymax(index);
        u0 = sqrt( 1 - v1*v1 );
        u1 = xmax(index);
        v0 = sqrt( 1 - u1*u1 );
        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        inArea(index) = inArea(index) - (u1-u0)*(v1-v0)/2 + arc;
    elseif topLeftIn % area[10]
        v1 = ymax(index);
        u0 = sqrt( 1 - v1*v1 );
        v0 = ymin(index);
        u1 = sqrt( 1 - v0*v0 );
        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        inArea(index) = ((u0+u1)/2 - xmin(index))*delY + arc;
    elseif botRightIn % area[01]
        u1 = xmax(index);
        v0 = sqrt( 1 - u1*u1 );
        u0 = xmin(index);
        v1 = sqrt( 1 - u0*u0 );
        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        inArea(index) = ((v0+v1)/2 - ymin(index))*delX + arc;
    else % area[00]
        u0 = xmin(index);
        v1 = sqrt( 1 - u0*u0 );
        v0 = ymin(index);
        u1 = sqrt( 1 - v0*v0 );
        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        inArea(index) = (u1-u0)*(v1-v0)/2 + arc;
    end
end

inArea = inArea * rX * rY;

if quadSym
    halfSize = size(inArea,1);
    inArea(1:halfSize,halfSize+1:halfSize*2) = fliplr(inArea);
    inArea(halfSize+1:halfSize*2,1:halfSize*2) = flipud( inArea(1:halfSize,1:halfSize*2));
end

end
