% classdef UtilTK
%   general purpose static functions
%
% July 3d, 2015
%   two functions, evalError and areSameVecs, are moved from cavityManager.
%
% August 12, 2014
%   mapped = mapping( x, y, mat, newX, newY ) added, which rotates mat(iy,ix) using the formula given in newX and newY
%
classdef UtilTK
    
    methods (Static)
        % error = sqrt( abs(Eold-Enew)^2 / abs(Enew)^2 )
        function err = evalError( Eold, Enew )
            Ediff = Enew - Eold;
            err = sqrt( sum(sum( conj(Ediff) .* Ediff )) / sum(sum( conj(Enew).*Enew )) );
        end
        
        function val = areSameVecs( v1, v2 )
            val = length(v1) == length(v2);
            if val
                val = all( v1 == v2 );
            end
        end
        
        function mapped = mapping( x, y, dat, newX, newY )
            % mapped = mapping( x, y, dat, newX, newY )
            %   calculate data at new coodrinate system or changing the data by rotating, moving or expanding,
            %   x and y are 2D coordinates in the original frame, dat(Ny,Nx) are 2D data
            %   and newX and newY are values in the new coordinate system
            %   or mapped( x, y ) = mat( newy, newx ), or newX/newY defines coordinate change, opposite to object change
            %   example :
            %       1) newX = cos * x - sin * y, newY = sin * x + cos * y
            %          This will rotate mat
            %       2) newX = 2*x, newY = 0.5*y
            %          This will shrink object by factor 2 in x direction and expand by factor 2 in y direction
            if min( size(x) ) == 1 && min( size(y) ) == 1
                [x, y] = meshgrid( x, y );
            end
            
            dx = x(1,2) - x(1,1);   xbase = x(1,1) - dx/2;  Nx = size(x,2);
            dy = y(2,1) - y(1,1);   ybase = y(1,1) - dy/2;  Ny = size(y,1);
            
            % for x(ix),y(iy), the value at newx(x(ix),y(iy)), newy(x(ix),y(iy)), is used
            if ischar( newX )
                newX = eval( newX );
            end
            newX = floor( (newX - xbase)/dx ) + 1;
            newX = min( max( newX, 1 ), Nx );
            if ischar( newY )
                newY = eval( newY );
            end
            newY = floor( (newY - ybase)/dy ) + 1;
            newY = min( max( newY, 1 ), Ny );
            
            mapped = interp2( 1:Nx, 1:Ny, dat, newX, newY );
        end
        
        function val = inputD( prompt, default )
            % val = inputD( prompt, default )
            % prompt and read one value, possibly with default
            val = input( prompt );
            if isempty( val ) && exist('default', 'var')
                val = default;
            end
        end
        
        function str = setVars( nameList, showIt, varargin )
            % str = setVars( prompt, nameList, varargin )
            %   find varargin{n} which is in a cell array, and create a string varargin{n} = varargin{n+1}
            in = 1;
            str = '';
            while in < length(varargin)
                if ischar( varargin{in} )
                    id = find( strcmp( varargin{in}, nameList ) == 1, 1 );
                    if ~isempty( id )
                        if ischar( varargin{in+1} )
                            rhs = sprintf('''%s''', varargin{in+1});
                        else
                            rhs = sprintf('[ %s ]', num2str( varargin{in+1} ));
                        end
                        oneCmd = sprintf( '%s = %s', nameList{id}, rhs);
                        if showIt
                            disp( oneCmd );
                        end
                        str = sprintf('%s %s;', str, oneCmd );
                        in = in + 1;
                    end
                end
                in = in + 1;
            end
        end
        
        function str = setVarsInputs( prompt, varargin )
            % usage : eval( UtilTK.setVarsInputs( 'set values', 'leng', 'RoC' ) ) or
            %         eval( UtilTK.setVarsInputs( 'leng' ) )
            % User is promprted to enter command, like leng = 4000; RoC = 1934
            % and this function returns a string which can be executed using eval
            str = '';
            if nargin == 0
                return;
            end
            if nargin > 2 % more than two variables
                if isempty( prompt )
                    prompt = 'Set values';
                end
                fprintf(1, '%s (name = val, ? for list, return to end)', prompt );
                promptMark = ' > ';
            else
                if nargin == 1
                    varargin{1} = prompt;
                    prompt = '';
                end
                fprintf(1, '%s %s = ', prompt, varargin{1} );
                promptMark = '';
            end
            % reapeat until empty input
            while true
                command = input( promptMark, 's');
                if isempty( command )
                    break;
                end
                % analyze one command, separated by ;, at a time, until all are analyzed
                while ~isempty(command)
                    % oneCmd + ';' + new command = old command
                    [oneCmd, command] = UtilTK.extactOne( command );
                    if isempty( oneCmd )
                        continue;
                    end
                    if oneCmd == '?'
                        for ivar = 1 : length(varargin)
                            fprintf(1, '''%s''  ', varargin{ivar});
                        end
                        fprintf(1, '\n');
                    else
                        ind = find( oneCmd == '=', 1 );
                        % if assignment, and if name list is given, make sure the name is in the list
                        if isempty(ind)
                            if length( varargin ) == 1
                                oneCmd = sprintf( '%s = %s', varargin{1}, oneCmd );
                            else
                                fprintf(1, 'command should be name = val\n');
                                continue;
                            end
                        else
                            name = strtrim( oneCmd(1:ind-1) );
                            if isempty( find( strcmp(varargin,name), 1 ) )
                                fprintf(1, 'unknown name : %s\n', name );
                                continue;
                            end
                        end
                        str = sprintf( '%s %s;', str, oneCmd );
                    end
                end
                if nargin <= 2
                    break;
                end
            end
        end
        
        function updated = updateArgList( origList, newDef, opt )
            % function updated = updateArgList( origList, newDef, opt )
            % newDef ( 'name1', var1, 'name2', var2 ) is used to update the original list.
            % opt = 0 : replace value if name exists, add new name and var if not
            % opt = 1 : only replace, no merge, and continue
            % opt = 2 : only replace, print warning
            updated = origList;
            if isempty( newDef )
                return;
            end
            firstWarning = true;
            for idnew = 1 : 2 : length( newDef )
                id = UtilTK.tagInVararg( newDef{idnew}, updated );
                if id == 0
                    if opt == 0
                        curLeng = length( updated );
                        updated{ curLeng : curLeng+1 } = newDef{idNew:idNew+1};
                    elseif opt == 2
                        fprintf(1,'>>> variable ''%s'' is not in the name list\n', newDef{idnew} );
                        if firstWarning
                            firstWarning = false;
                            fprintf(1,'    names are :');
                            for id = 1 : 2 : length(updated)
                                fprintf(1,'''%s''\t', updated{id});
                            end
                            fprintf(1,'\n');
                        end
                    end
                else
                    updated{id+1} = newDef{idnew+1};
                end
            end
        end
        
        function [oneCmd, command] = extactOne( command )
            % [oneCmd, command] = extactOne( command )
            % extract one command separated by ; and return in oneCmd
            % command is the original with the oneCmd trimmed out
            separator = find( command == ';', 1 );
            if isempty( separator )
                oneCmd = command;
                command = '';
            else
                oneCmd = command(1:separator-1);
                command = command(separator+1:end);
            end
            oneCmd = strtrim( oneCmd );
            command = strtrim( command );
        end
        
        function id = namesInList( name, tagList )
            % find tag(s) in tagList which match with name.
            %    name = 'T', tagList = {'T', 'Thick'}
            %
            % If multiple tags match, like name = 'T' and tag ={'T', Thick'}
            % full match, 'T', is chosen.
            % If tagList = {'Thick', 'Tilt'}, both are returned.
            
            id = find( strncmp( name, tagList, length(name) ) );
            if length( id ) > 1
                for idu = id
                    if length( tagList{idu} ) == length(name)
                        id = idu;
                        break;
                    end
                end
            end
        end
        
        function id = oneNameInList( name, tagList )
            % find the unique tag ID using namesInList
            %    name = 'T', tagList = {'T', 'Thick'}
            %
            % If there is no match, return id = 0.
            % If there are multiple mathces, an error message is printed and exits
            id = UtilTK.namesInList( name, tagList );
            if isempty(id)
                id = 0;
            elseif length( id ) > 1
                str = sprintf('''%s''', tagList{id(1)});
                for idu = id(2:end)
                    if idu == id(end)
                        str = sprintf('%s and ''%s''', str, tagList{idu} );
                    else
                        str = sprintf('%s, ''%s''', str, tagList{idu} );
                    end
                end
                error( 'name ''%s'' matches with following candidates {%s}', name, str );
            end
        end
        
        function newList = idenfyTags( tagList, varargin )
            % vararg is valid argments to specify parameters, pass validArgList.
            % tag names in vararg is replaced by tags in tagList.
            % names in vararg are replaced by the full names in tagList.
            % If there are any names which are not in tagList, return empty newList
            
            if ~UtilTK.validArgList( varargin{:} )
                error('argument needs to be a list of ''name'' and value pairs');
            end
            newList = varargin;
            for id = 1 : 2: length( newList )
                tagID = UtilTK.oneNameInList( newList{id}, tagList );
                if tagID == 0
                    newList = {};
                    return;
                end
                newList{id} = tagList{tagID};
            end
        end
        
        function [val, inputs] = extractVal( tag, inputs, defaultVal )
            % [val, inputs] = extractVal( tag, inputs, defaultVal )
            % for the input string, 'name1', val1, 'name2', val2, ...
            % find a 'nameX' which is in the tag list and returns the
            % associtaed value, valN.
            % When a nameX is found, the name section of the return value "inputs" is emptied
            id = UtilTK.tagInVararg( tag, inputs );
            if id == 0
                if exist( 'defaultVal', 'var' )
                    val = defaultVal;
                else
                    val = [];
                end
            else
                val = inputs{id+1};
                if nargout == 2
                    inputs{id} = '';
                end
            end
        end
        
        function updateProperties( object, tagList, inputs )
            % object.(tagName) = tagged value in vararg
            
            for tagID = 1 : length( tagList )
                id = UtilTK.tagInVararg( tagList(tagID), inputs );
                if id > 0
                    object.(tagList{tagID}) = inputs{ id + 1 };
                end
            end
        end
        
        
        function id = tagInVararg( tag, inputs )
            % id = tagInVararg( tag, inputs )
            % find tag names in vararg name section ( 'name1', val1, 'name2', val2, etc )
            
            id = find( strcmp( inputs(1:2:end), tag ), 1 );
            if isempty(id)
                id = 0;
            else
                id = id * 2 - 1;
            end
        end
        
        function val = validArgList( varargin )
            % return true if argarg is 'tag1', val1, 'tag2', vale2 ...
            val = false;
            if mod(length( varargin ),2) == 0
                % needs to be even
                for itag = 1 : 2 : length(varargin)-1
                    if ~ischar( varargin{itag} )
                        return;
                    end
                end
                val = true;
            end
        end
        
        % sumIn and sumOut need to be improved to handle the corner of a square grid
        function val = sumIn( xlist, ampList, rx, ry, x0, y0 )
            % function val = sumIn( xlist, ampList, rx, ry, x0, y0 )
            % sum(sum( ampList.^2) ) * dx^2 within radius (rx,ry) and offcentered at (x0,y0)
            
            if ~exist('rx','var')
                rx = max(abs(xlist));
            end
            if ~exist('ry','var')
                ry = rx;
            end
            if ~exist('x0','var')
                x0 = 0;
            end
            if ~exist('y0','var')
                y0 = 0;
            end
            
            % leng = length( xlist );
            datSize = min( length(rx), length(ry) );
            val = zeros( 1, datSize );
            for jj = 1 : datSize
                val(jj) = ...
                    sum( abs(ampList( UtilTK.Rsquare( xlist, x0, y0, rx(jj), ry(jj) ) <= 1 )).^2 ) * ...
                    (xlist(2)-xlist(1)).^2;
            end
        end
        
        function val = sumOut( xlist, ampList, rx, ry, x0, y0 )
            % function val = sumOut( xlist, ampList, rx, ry, x0, y0 )
            % sum(sum( ampList.^2) ) * dx^2 outside of radius (rx,ry) and offcentered at (x0,y0)
            
            if ~exist('rx','var')
                rx = max(abs(xlist));
            end
            if ~exist('ry','var')
                ry = rx;
            end
            if ~exist('x0','var')
                x0 = 0;
            end
            if ~exist('y0','var')
                y0 = 0;
            end
            
            % leng = length( xlist );
            datSize = min( length(rx), length(ry) );
            val = zeros( 1, datSize );
            for jj = 1 : datSize
                val(jj) = ...
                    sum( abs(ampList( UtilTK.Rsquare( xlist, x0, y0, rx(jj), ry(jj) ) > 1 )).^2 ) * ...
                    (xlist(2)-xlist(1)).^2;
            end
        end
        
        function r2 = Rsquare( x, x0, y0, rx, ry )
            % function r2 = Rsquare( x, x0, y0, rx, ry )
            % (x-x0)^2 / rx^2 + (y-y0)^2 / ry^2
            
            if ~exist('rx','var')
                rx = 1;
            end
            if ~exist('ry','var')
                ry = 1;
            end
            if ~exist('x0','var')
                x0 = 0;
            end
            if ~exist('y0','var')
                y0 = 0;
            end
            
            if size(x,2) == 1
                x = x';
            end
            
            [xg,yg] = meshgrid( x, x );
            r2 = ( xg - x0 ).^2 / rx^2 + (yg - y0).^2 / ry^2;
            
        end
        
        function mat1 = useQuarter( mat4 )
            % mat1 = useQuarter( mat4 )
            % return the central (N/2xN/2) area
            Nlarge = length( mat4 );
            Nsmall = Nlarge/2;
            mat1 = mat4( 1+(Nlarge-Nsmall)/2 : (Nlarge+Nsmall)/2, ...
                1+(Nlarge-Nsmall)/2 : (Nlarge+Nsmall)/2 );
        end
        
        function mat4 = expand2x2( mat1, Nlarge )
            % mat4 = expand2x2( mat1, Nlarge )
            % expand mat1 to NlargexNlarge by filling zeros outside of mat1
            % default Nlarge is 2 x mat1 size
            Nsmall = length( mat1 );
            if ~exist('Nlarge', 'var')
                Nlarge = Nsmall * 2;
            end
            mat4 = zeros( Nlarge, Nlarge );
            mat4( 1+(Nlarge-Nsmall)/2 : (Nlarge+Nsmall)/2, ...
                1+(Nlarge-Nsmall)/2 : (Nlarge+Nsmall)/2 ) = mat1;
        end
        
        function filter = lowpassFilter( Nx, Ny, Fcutoff, Nfilter )
            % filter = lowpassFilter( Nx, Ny, Fcutoff, Nfilter )
            % simple low pass filter
            if Fcutoff <= 0 || Fcutoff > 0.5
                error( 'Fcutoff (%g) must be between (0, 0.5]', Fcutoff );
            end
            if mod(Nfilter,2) ~= 0
                error( 'Nfilter (%g) must be an integer', Nfilter );
            end
            [fx,fy] = meshgrid( ((1:Nx) - fix(Nx/2)-1)/Nx, ((1:Ny) - fix(Ny/2)-1)/Ny );
            filter = 1 ./ (1 + ((fx.^2 + fy.^2) / Fcutoff^2 ).^Nfilter );
        end
        
        function mapOut = lowPassMap( mapIn, xIn, yIn, xOut, yOut, Fcutoff, Nfilter, tilt, cutHF )
            % function mapOut = lowPassMap( mapIn, xIn, yIn, xOut, yOut, Fcutoff[0.5], Nfilter[20], tilt[0], cutHF[false] )
            %
            % mapFiltered = real( ifft2(ifftshift( filterWgt .* fftshift(fft2(mapIn)))) );
            % mapOut = cos(tilt) * interp2( xIn*cos(tilt), mapFiltered, xOut )
            %
            % apply low-pass butterworth filter on input map, defined in [xIn, yIn]
            % and returns map in [xOut, yOut]
            % cutoff frequency (0~0.5) and order of filter can be specified
            %     filter = 1 ./ (1.0 + (fx^2 + yf^2) ./ Fcutoff^2 ).^(Nfiter))
            %
            % lowpass filter is applied when the new grid size is larger than factor 2
            %
            % tilt is applied so that this map is to be seen by a field coming with angle tilt (radian)
            % i.e., when mapping from [xIn,yIn] to [xOut,yOut], actually xIn*cos(tilt) is used and
            % the mapOut = mapOut*cos(tilt)
            %
            % if cutHF is true, high frequecy component is cut down as :
            % mapF = filterWgt .* fftshift(fft2(mapIn)) is shrink down to the size of (xOut,yOut)
            % which cuts the higest frequency by size of Xout / size of Xin = Fmax out / Fmax in
            %
            
            if size(mapIn,2) ~= length(xIn) || size(mapIn,1) ~= length(yIn)
                error( 'mapIn size ( [%d, %d] ) must match with the size of xIn and yIn', ...
                    size(mapIn), length(xIn), length(yIn) );
            end
            
            % if x and y are in one row or column vector, make x a row vector and y a column vector
            if size(xOut,2) == 1
                xOut = xOut';
            end
            if size(yOut,1) == 1
                yOut = yOut';
            end
            if size(xIn,2) == 1
                xIn = xIn';
            end
            if size(yIn,1) == 1
                yIn = yIn';
            end

            if ~exist( 'Fcutoff', 'var' ) || Fcutoff == 0
                Fcutoff = 0.5;
            end
            if ~exist( 'Nfilter', 'var' ) || Nfilter == 0
                Nfilter = 20;
            end
            if ~exist( 'tilt', 'var' )
                tilt = 0;
            end
            if ~exist( 'cutHF', 'var' )
                cutHF = false;
            end
            
            binRatio = min(xIn(1,2)-xIn(1,1),yIn(2,1)-yIn(1,1))/max(xOut(1,2)-xOut(1,1),yOut(2,1)-yOut(1,1));
            
            if binRatio < 0.5
                
                Nfilter = fix( Nfilter );
                
                % apply filter
                NaN_Mask = ~isfinite(mapIn);
                mapIn(NaN_Mask)=0;
                mapTemp = fftshift(fft2(mapIn));
                
                % cut the high frequency component and use the coarse spacing xOut and yOut
                if cutHF
                    NXin = length(xIn);
                    NYin = length(yIn);
                
                    % NXout is number of xOut points in the range of xIn, like W = 40cm and mirror data is in 30cm
                    NXout = length( find( xOut(1,:)) >= xIn(1,1) & xOut(1,:) <= xIn(1,end) );
                    NYout = length( find( yOut(:,1)) >= yIn(1,1) & yOut(:,1) <= yIn(end,1) );

                    % When the frequency range size is cut from Nin to Nout without changing the frequency spacing,
                    % the window size W is unchanged, while the grid spacing becomes larger by Nin/Nout.
                    % Out spacing = W/Nin * (Nin/Nout) = W/Nout = spacing of xOut
                    mapTemp = mapTemp( 1+floor((NYin-NYout)/2):floor((NYin+NYout)/2),...
                        1+floor((NXin-NXout)/2):floor((NXin+NXout)/2)) * (NXout*NYout)/(NXin*NYin);
                    
                    xIn = (xIn(1,end)-xIn(1,1))/(2*NXout)*(-NXout+1:2:NXout-1);
                    yIn = (yIn(end,1)-yIn(1,1))/(2*NYout)*(-NYout+1:2:NYout-1);
                end
                                % Anti-aliasing filter;
                % calculate filter
                Fcutoff = min( [0.5*binRatio, Fcutoff, 0.5] );
                filterWgt = UtilTK.lowpassFilter( length(xIn), length(yIn), Fcutoff, Nfilter );

                mapIn = real( ifft2 ( ifftshift(filterWgt .* mapTemp) ) );
                if ~cutHF
                    % for frequency cut case, how to restore the NaN information ???
                    mapIn(NaN_Mask) = NaN;
                end
            end
            
            if size(yIn,1) == 1
                yIn = yIn';
            end
            [xIn, yIn] = meshgrid( xIn(1,:)*cos(tilt), yIn(:,1) );
            mapOut = cos(tilt) * interp2(xIn,yIn,mapIn,xOut,yOut);
        end
    end
end

