function [leng, loss, power] = runLG( pList, arm )
% function [leng, loss, power] = runLG( pList, arm )
% calculate resonances in a FP cavity with different input LG(pList,0), using mirrors of LLO X or Y arm.
% for each input, the following four cases are analyzed
%   60cm A, no map
%   32.6cm A, no map
%   32.6cm A, ITM map
%   32.6cm A, ETM/ETM map
% E.g., loss(1:4,1) are lesses for the 4 cases above with the input mode LG(pList(1),0)
% arm = 'X' (or 'Y') to calculate X or Y arm
if ~exist('arm')
    arm = 'Y';
end

leng = zeros(4,length(pList));
loss = zeros(4,length(pList));
power = zeros(4,length(pList));

largeA = 0.6;
nominalA = 0.326;

for ip = 1:length( pList )
    % large aperture case without maps
    fp = FPIFOv( 'extraArmLoss', 0, 'fieldMode', 'LG', 'ind1', pList(ip), 'withMapI', false, 'withMapE', false, 'Aperture', largeA, 'arm', arm );
    [leng(1,ip), loss(1,ip), power(1,ip)] = analOne( fp );
    delete(fp);
    % nominal aperture without maps
    fp = FPIFOv( 'extraArmLoss', 0, 'fieldMode', 'LG', 'ind1', pList(ip), 'withMapI', false, 'withMapE', false, 'Aperture', nominalA, 'arm', arm );
    [leng(2,ip), loss(2,ip), power(2,ip)] = analOne( fp );
    delete(fp);
    % nominal aperture with ITM map
    fp = FPIFOv( 'extraArmLoss', 0, 'fieldMode', 'LG', 'ind1', pList(ip), 'withMapI', true, 'withMapE', false, 'Aperture', nominalA, 'arm', arm );
    [leng(3,ip), loss(3,ip), power(3,ip)] = analOne( fp );
    delete(fp);
    % mominal aperture with ITM and ETM maps
    fp = FPIFOv( 'extraArmLoss', 0, 'fieldMode', 'LG', 'ind1', pList(ip), 'withMapI', true, 'withMapE', true, 'Aperture', nominalA, 'arm', arm );
    [leng(4,ip), loss(4,ip), power(4,ip)] = analOne( fp );
    delete(fp);
end
end

function [leng, loss, power] = analOne( fp )
fp.lock(1e-6);
 [~,leng] = fp.getPropLeng('FPprop');  % micro adjustment of the progapator between ITM and ETM
 loss = fp.roundtripLoss('ITM-HR-o');  % roundtrip loss, starting from outgoing field from ITM-HR
 power = fp.power('ETM-HR-o');         % power of the out going field from ETM-HR
end